# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tasks related to I/O."""

import base64
import io
import os
import shutil
import zipfile

from six.moves import urllib

from gcloud import storage

from artman.tasks import task_base
from artman.utils.logger import logger


class PrepareGoogleapisDirTask(task_base.TaskBase):

    default_provides = ('remote_repo_dir')

    def execute(self, root_dir, files_dict={}):
        repo_root = os.path.abspath(os.path.join(root_dir, os.pardir))
        if os.path.exists(os.path.realpath(os.path.expanduser(repo_root))):
            # Do nothing if the repo_root exists. The repo_root exists if
            # artman is running locally.
            return
        logger.info('root repo: %s' % repo_root)
        try:
            os.makedirs(repo_root)
        except OSError as e:
            raise e
        testfile = urllib.request.FancyURLopener()
        tmp_repo_file = os.path.join(repo_root, "file.zip")
        testfile.retrieve(
            "https://github.com/googleapis/googleapis/archive/master.zip",
            tmp_repo_file)
        zip_ref = zipfile.ZipFile(tmp_repo_file, 'r')
        zip_ref.extractall(repo_root)
        zip_ref.close()
        os.remove(tmp_repo_file)
        shutil.move(os.path.join(repo_root, "googleapis-master"),
                    os.path.join(repo_root, "googleapis"))
        remote_repo_dir = os.path.join(repo_root, "googleapis")
        # Write/overwrite the additonal files into the remote_repo_dir so that
        # user can include additional files which are not in the public repo.
        for f, content in files_dict.items():
            filename = os.path.join(remote_repo_dir, f)
            if not os.path.exists(os.path.dirname(filename)):
                os.makedirs(os.path.dirname(filename))
            with io.open(filename, "w+", encoding='UTF-8') as text_file:
                text_file.write(base64.b64decode(content))
        return remote_repo_dir


class PrepareOutputDirectoryTask(task_base.TaskBase):
    def execute(self, output_dir):
        self.exec_command(['mkdir', '-p', output_dir])



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tasks related to generation of GAPIC wrappers"""

import os
import glob
from ruamel import yaml

from artman.tasks import task_base
from artman.utils import task_utils


class GapicConfigGenTask(task_base.TaskBase):
    """Generates GAPIC config file"""
    default_provides = 'gapic_config_path'

    def execute(self, toolkit_path, descriptor_set, service_yaml,
                output_dir, api_name, api_version, organization_name):
        api_full_name = task_utils.api_full_name(
            api_name, api_version, organization_name)
        config_gen_dir = os.path.join(
            output_dir, api_full_name + '-config-gen')
        self.exec_command(['mkdir', '-p', config_gen_dir])
        config_gen_path = os.path.join(config_gen_dir,
                                       api_full_name + '_gapic.yaml')
        args = [
            '--descriptor_set=' + os.path.abspath(descriptor_set),
            '--output=' + os.path.abspath(config_gen_path),
        ]
        if service_yaml:
            args = args + ['--service_yaml=' + os.path.abspath(service_yaml)]
        self.exec_command(
            task_utils.gapic_gen_task(toolkit_path, ['GAPIC_CONFIG'] + args))

        return config_gen_path


class DiscoGapicConfigGenTask(task_base.TaskBase):
    """Generates GAPIC config file from a Discovery document"""
    default_provides = 'gapic_config_path'

    def execute(self, toolkit_path, discovery_doc,
        output_dir, api_name, api_version, organization_name):
        api_full_name = task_utils.api_full_name(
            api_name, api_version, organization_name)
        config_gen_dir = os.path.join(
            output_dir, api_full_name + '-config-gen')
        self.exec_command(['mkdir', '-p', config_gen_dir])
        config_gen_path = os.path.join(config_gen_dir,
                                       api_full_name + '_gapic.yaml')
        args = [
            '--discovery_doc=' + os.path.abspath(
                os.path.expanduser(discovery_doc)),
            '--output=' + os.path.abspath(config_gen_path)
        ]
        self.exec_command(
            task_utils.gapic_gen_task(toolkit_path, ['DISCOGAPIC_CONFIG'] + args))

        return config_gen_path


class GapicConfigMoveTask(task_base.TaskBase):
    """Move config file to gapic_yaml location"""

    def _move_to(self, gapic_config_path, gapic_yaml):
        if not gapic_yaml:
            raise ValueError('Could not move generated config file ' \
                'from "{0}" to destination": No location specified'.format(
                os.path.abspath(gapic_config_path)))
        conf_out = os.path.abspath(gapic_yaml)
        if os.path.exists(conf_out):
            # TODO (issue #80): no need to test in remote environment
            olderVersion = conf_out + '.old'
            print('File already exists, save the old version as ' + olderVersion)
            self.exec_command(['mv', conf_out, olderVersion])
        return conf_out

    def execute(self, gapic_config_path, gapic_yaml):
        conf_out = self._move_to(gapic_config_path, gapic_yaml)
        self.exec_command(['mkdir', '-p', os.path.dirname(conf_out)])
        self.exec_command(['cp', gapic_config_path, conf_out])
        return

    def validate(self):
        return []


class GapicCodeGenTask(task_base.TaskBase):
    """Generates GAPIC wrappers"""
    default_provides = 'gapic_code_dir'

    def execute(self, language, toolkit_path, descriptor_set, service_yaml,
                gapic_yaml, package_metadata_yaml, proto_package,
                gapic_code_dir, api_name, api_version, organization_name,
                aspect, samples, generator_args, grpc_service_config=''):
        existing = glob.glob('%s/*' % gapic_code_dir)
        if existing:
            self.exec_command(['rm', '-r'] + existing)
        gapic_args = []
        if proto_package:
            gapic_args.append('--package=' + proto_package)
        if gapic_yaml:
            gapic_args.append('--gapic_yaml=' + os.path.abspath(gapic_yaml))
        if samples:
            sample_yamls = self._get_sample_yamls(samples)
            if sample_yamls:
                gapic_args.append('--sample_yamls')
                for sample_yaml in sample_yamls:
                    gapic_args.append(sample_yaml)
        if grpc_service_config:
            gapic_args.append('--grpc_service_config=' + os.path.abspath(grpc_service_config))

        args = [
            '--descriptor_set=' + os.path.abspath(descriptor_set),
            '--package_yaml2=' + os.path.abspath(package_metadata_yaml),
            '--output=' + os.path.abspath(gapic_code_dir),
            '--language=' + language,
        ]
        if service_yaml:
            args = args + ['--service_yaml=' + os.path.abspath(service_yaml)]
        args = args + gapic_args

        if generator_args:
          args = args + generator_args.split(' ')

        gapic_artifact = ''
        if aspect == 'ALL':
            gapic_artifact = 'LEGACY_GAPIC_AND_PACKAGE'
        elif aspect == 'CODE':
            gapic_artifact = 'GAPIC_CODE'
        elif aspect == 'PACKAGE':
            gapic_artifact = 'GAPIC_PACKAGE'
        else:
            raise ValueError('GapicCodeGenTask: no generation turned on')

        self.exec_command(
            task_utils.gapic_gen_task(toolkit_path, [gapic_artifact] + args))

        return gapic_code_dir

    def _get_sample_yamls(self, path):
        sample_path = os.path.abspath(path)
        if os.path.isfile(sample_path):
            return [sample_path]
        elif os.path.isdir(sample_path):
            sample_paths = []
            for root, dirnames, filenames in os.walk(sample_path):
                for filename in filenames:
                    if filename.endswith('.yaml'):
                        sample_paths.append(os.path.join(root, filename))
            return sample_paths
        else:
            raise ValueError('GapicCodeGenTask: `--samples` directory or file does not exist.')

class DiscoGapicCodeGenTask(task_base.TaskBase):
    """Generates GAPIC wrappers from a Discovery document"""
    default_provides = 'gapic_code_dir'

    def execute(self, language, toolkit_path, discovery_doc,
        gapic_yaml, package_metadata_yaml,
        gapic_code_dir, api_name, api_version, organization_name, root_dir):
        existing = glob.glob('%s/*' % gapic_code_dir)
        if existing:
            self.exec_command(['rm', '-r'] + existing)
        gapic_args = ['--gapic_yaml=' + os.path.abspath(gapic_yaml)]
        args = [
                   '--discovery_doc=' + os.path.join(root_dir, discovery_doc),
                   '--package_yaml2=' + os.path.abspath(package_metadata_yaml),
                   '--output=' + os.path.abspath(gapic_code_dir),
                   '--language=' + language,
                   ] + gapic_args

        self.exec_command(
            task_utils.gapic_gen_task(toolkit_path, ['LEGACY_DISCOGAPIC_AND_PACKAGE'] + args))

        return gapic_code_dir


class CSharpGapicPackagingTask(task_base.TaskBase):
    def execute(self, gapic_code_dir, grpc_code_dir, proto_code_dir, gapic_yaml):
        with open(gapic_yaml) as f:
            gapic_config = yaml.load(f, Loader=yaml.Loader)
        package_name = gapic_config.get('language_settings').get('csharp').get('package_name')
        package_root = '{0}/{1}'.format(gapic_code_dir, package_name)
        prod_dir = '{0}/{1}'.format(package_root, package_name)
        # Copy proto/grpc .cs files into prod directory
        self.exec_command(['sh', '-c', 'cp {0}/*.cs {1}'.format(proto_code_dir, prod_dir)])
        self.exec_command(['sh', '-c', 'cp {0}/*.cs {1}'.format(grpc_code_dir, prod_dir)])



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A canonical example of a flow class supported by gapic pipeline."""

from artman.pipelines import pipeline_base
from artman.tasks import sample_tasks
from taskflow.patterns import linear_flow


class SamplePipeline(pipeline_base.PipelineBase):

    def __init__(self, **kwargs):
        super(SamplePipeline, self).__init__(**kwargs)

    def do_build_flow(self, **kwargs):
        sleep_secs = kwargs.get('sleep_secs')
        flow = linear_flow.Flow('sample-flow')
        flow.add(sample_tasks.SampleTask('SampleTask',
                                         inject={'sleep_secs': sleep_secs}))
        return flow

    def validate_kwargs(self, **kwargs):
        if 'sleep_secs' not in kwargs:
            raise ValueError('sleep_secs must be provided')



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: user_config.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='user_config.proto',
  package='googleapis.artman',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x11user_config.proto\x12\x11googleapis.artman\"l\n\nUserConfig\x12-\n\x05local\x18\x01 \x01(\x0b\x32\x1e.googleapis.artman.LocalConfig\x12/\n\x06github\x18\x02 \x01(\x0b\x32\x1f.googleapis.artman.GitHubConfig\"\x1e\n\x0bLocalConfig\x12\x0f\n\x07toolkit\x18\x01 \x01(\t\"/\n\x0cGitHubConfig\x12\x10\n\x08username\x18\x01 \x01(\t\x12\r\n\x05token\x18\x02 \x01(\tb\x06proto3')
)




_USERCONFIG = _descriptor.Descriptor(
  name='UserConfig',
  full_name='googleapis.artman.UserConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='local', full_name='googleapis.artman.UserConfig.local', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='github', full_name='googleapis.artman.UserConfig.github', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=40,
  serialized_end=148,
)


_LOCALCONFIG = _descriptor.Descriptor(
  name='LocalConfig',
  full_name='googleapis.artman.LocalConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='toolkit', full_name='googleapis.artman.LocalConfig.toolkit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=150,
  serialized_end=180,
)


_GITHUBCONFIG = _descriptor.Descriptor(
  name='GitHubConfig',
  full_name='googleapis.artman.GitHubConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='username', full_name='googleapis.artman.GitHubConfig.username', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='token', full_name='googleapis.artman.GitHubConfig.token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=182,
  serialized_end=229,
)

_USERCONFIG.fields_by_name['local'].message_type = _LOCALCONFIG
_USERCONFIG.fields_by_name['github'].message_type = _GITHUBCONFIG
DESCRIPTOR.message_types_by_name['UserConfig'] = _USERCONFIG
DESCRIPTOR.message_types_by_name['LocalConfig'] = _LOCALCONFIG
DESCRIPTOR.message_types_by_name['GitHubConfig'] = _GITHUBCONFIG
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

UserConfig = _reflection.GeneratedProtocolMessageType('UserConfig', (_message.Message,), dict(
  DESCRIPTOR = _USERCONFIG,
  __module__ = 'user_config_pb2'
  # @@protoc_insertion_point(class_scope:googleapis.artman.UserConfig)
  ))
_sym_db.RegisterMessage(UserConfig)

LocalConfig = _reflection.GeneratedProtocolMessageType('LocalConfig', (_message.Message,), dict(
  DESCRIPTOR = _LOCALCONFIG,
  __module__ = 'user_config_pb2'
  # @@protoc_insertion_point(class_scope:googleapis.artman.LocalConfig)
  ))
_sym_db.RegisterMessage(LocalConfig)

GitHubConfig = _reflection.GeneratedProtocolMessageType('GitHubConfig', (_message.Message,), dict(
  DESCRIPTOR = _GITHUBCONFIG,
  __module__ = 'user_config_pb2'
  # @@protoc_insertion_point(class_scope:googleapis.artman.GitHubConfig)
  ))
_sym_db.RegisterMessage(GitHubConfig)


# @@protoc_insertion_point(module_scope)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pipelines that run protoc core codegen. The generated core library for each
language contains the well known types, defined by protobuf, for that language.
"""

from artman.pipelines import code_generation as code_gen
from artman.tasks import protoc_tasks
from artman.tasks import package_metadata_tasks
from artman.utils import task_utils


class CoreProtoPipeline(code_gen.CodeGenerationPipelineBase):

    def __init__(self, **kwargs):
        super(CoreProtoPipeline, self).__init__(
            get_core_task_factory(kwargs['language']), **kwargs)


class CoreTaskFactoryBase(code_gen.TaskFactoryBase):

    def get_tasks(self, **kwargs):
        return task_utils.instantiate_tasks(
            self._get_core_codegen_tasks(**kwargs), kwargs)

    def _get_core_codegen_tasks(self, **kwargs):
        raise NotImplementedError('Subclass must implement abstract method')

    def get_validate_kwargs(self):
        return code_gen.COMMON_REQUIRED

    def get_invalid_kwargs(self):
        return []


class _GoCoreTaskFactory(CoreTaskFactoryBase):
    """Responsible for the protobuf flow for Go language."""

    def _get_core_codegen_tasks(self, **kwargs):
        return [
            protoc_tasks.ProtoCodeGenTask,
            protoc_tasks.GoCopyTask,
        ]

    def get_validate_kwargs(self):
        return ['gapic_yaml', 'gapic_code_dir'] + code_gen.COMMON_REQUIRED


class _CSharpCoreTaskFactory(CoreTaskFactoryBase):

    def _get_core_codegen_tasks(self, **kwargs):
        return [protoc_tasks.ProtoCodeGenTask]


class _JavaCoreTaskFactory(CoreTaskFactoryBase):
    """Responsible for the core protobuf flow for Java language."""

    def _get_core_codegen_tasks(self, **kwargs):
        return [protoc_tasks.ProtoDescGenTask,
                protoc_tasks.ProtoCodeGenTask,
                protoc_tasks.ResourceNameGenTask,
                package_metadata_tasks.PackageMetadataConfigGenTask,
                package_metadata_tasks.ProtoPackageMetadataGenTask,
                protoc_tasks.JavaProtoCopyTask]


_CORE_TASK_FACTORY_DICT = {
    'go': _GoCoreTaskFactory,
    'csharp': _CSharpCoreTaskFactory,
    'java': _JavaCoreTaskFactory,
}


def get_core_task_factory(language):
    cls = _CORE_TASK_FACTORY_DICT.get(language)
    if cls:
        return cls()
    else:
        raise ValueError('No core task factory found for language: '
                         + language)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: config.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='config.proto',
  package='googleapis.artman',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x0c\x63onfig.proto\x12\x11googleapis.artman\"e\n\x06\x43onfig\x12+\n\x06\x63ommon\x18\x01 \x01(\x0b\x32\x1b.googleapis.artman.Artifact\x12.\n\tartifacts\x18\x02 \x03(\x0b\x32\x1b.googleapis.artman.Artifact\"\xc5\x0c\n\x08\x41rtifact\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x61pi_name\x18\x02 \x01(\t\x12\x13\n\x0b\x61pi_version\x18\x03 \x01(\t\x12\x19\n\x11organization_name\x18\x04 \x01(\t\x12?\n\rrelease_level\x18\x05 \x01(\x0e\x32(.googleapis.artman.Artifact.ReleaseLevel\x12\x17\n\x0fsrc_proto_paths\x18\x06 \x03(\t\x12?\n\nproto_deps\x18\x07 \x03(\x0b\x32+.googleapis.artman.Artifact.ProtoDependency\x12\x44\n\x0ftest_proto_deps\x18\x08 \x03(\x0b\x32+.googleapis.artman.Artifact.ProtoDependency\x12\x14\n\x0cservice_yaml\x18\t \x01(\t\x12\x12\n\ngapic_yaml\x18\n \x01(\t\x12\x15\n\rproto_package\x18\x12 \x01(\t\x12\x0f\n\x07samples\x18\x13 \x01(\t\x12\x1b\n\x13grpc_service_config\x18\x14 \x01(\t\x12.\n\x04type\x18\x0c \x01(\x0e\x32 .googleapis.artman.Artifact.Type\x12\x36\n\x08language\x18\r \x01(\x0e\x32$.googleapis.artman.Artifact.Language\x12\x43\n\x0fpackage_version\x18\x0e \x01(\x0b\x32*.googleapis.artman.Artifact.PackageVersion\x12\x42\n\x0fpublish_targets\x18\x0f \x03(\x0b\x32).googleapis.artman.Artifact.PublishTarget\x12\x15\n\rdiscovery_doc\x18\x10 \x01(\t\x12\x32\n\x06\x61spect\x18\x11 \x01(\x0e\x32\".googleapis.artman.Artifact.Aspect\x12\x1d\n\x15language_out_override\x18\x15 \x01(\t\x1a\x33\n\x0fProtoDependency\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x12\n\nproto_path\x18\x02 \x01(\t\x1a]\n\x0ePackageVersion\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x1c\n\x14grpc_dep_lower_bound\x18\x02 \x01(\t\x12\x1c\n\x14grpc_dep_upper_bound\x18\x03 \x01(\t\x1a\xb7\x02\n\rPublishTarget\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08location\x18\x02 \x01(\t\x12<\n\x04type\x18\x03 \x01(\x0e\x32..googleapis.artman.Artifact.PublishTarget.Type\x12V\n\x12\x64irectory_mappings\x18\x04 \x03(\x0b\x32:.googleapis.artman.Artifact.PublishTarget.DirectoryMapping\x1a;\n\x10\x44irectoryMapping\x12\x0b\n\x03src\x18\x01 \x01(\t\x12\x0c\n\x04\x64\x65st\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\"3\n\x04Type\x12\x1f\n\x1bPUBLISHING_TYPE_UNSPECIFIED\x10\x00\x12\n\n\x06GITHUB\x10\x01\"J\n\x0cReleaseLevel\x12\x1d\n\x19RELEASE_LEVEL_UNSPECIFIED\x10\x00\x12\x06\n\x02GA\x10\x01\x12\x08\n\x04\x42\x45TA\x10\x02\x12\t\n\x05\x41LPHA\x10\x03\"r\n\x04Type\x12\t\n\x05GAPIC\x10\x00\x12\x10\n\x0cGAPIC_CONFIG\x10\x01\x12\x08\n\x04GRPC\x10\x02\x12\x0c\n\x08PROTOBUF\x10\x03\x12\x0e\n\nDISCOGAPIC\x10\x04\x12\x15\n\x11\x44ISCOGAPIC_CONFIG\x10\x05\x12\x0e\n\nGAPIC_ONLY\x10\x63\"m\n\x08Language\x12\x18\n\x14LANGUAGE_UNSPECIFIED\x10\x00\x12\x08\n\x04JAVA\x10\x01\x12\n\n\x06PYTHON\x10\x02\x12\n\n\x06NODEJS\x10\x03\x12\x06\n\x02GO\x10\x04\x12\x07\n\x03PHP\x10\x05\x12\n\n\x06\x43SHARP\x10\x06\x12\x08\n\x04RUBY\x10\x07\"(\n\x06\x41spect\x12\x07\n\x03\x41LL\x10\x00\x12\x08\n\x04\x43ODE\x10\x01\x12\x0b\n\x07PACKAGE\x10\x02J\x04\x08\x0b\x10\x0cR\x11import_proto_pathb\x06proto3')
)



_ARTIFACT_PUBLISHTARGET_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='googleapis.artman.Artifact.PublishTarget.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PUBLISHING_TYPE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GITHUB', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1323,
  serialized_end=1374,
)
_sym_db.RegisterEnumDescriptor(_ARTIFACT_PUBLISHTARGET_TYPE)

_ARTIFACT_RELEASELEVEL = _descriptor.EnumDescriptor(
  name='ReleaseLevel',
  full_name='googleapis.artman.Artifact.ReleaseLevel',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='RELEASE_LEVEL_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GA', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BETA', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ALPHA', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1376,
  serialized_end=1450,
)
_sym_db.RegisterEnumDescriptor(_ARTIFACT_RELEASELEVEL)

_ARTIFACT_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='googleapis.artman.Artifact.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='GAPIC', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GAPIC_CONFIG', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GRPC', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PROTOBUF', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DISCOGAPIC', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DISCOGAPIC_CONFIG', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GAPIC_ONLY', index=6, number=99,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1452,
  serialized_end=1566,
)
_sym_db.RegisterEnumDescriptor(_ARTIFACT_TYPE)

_ARTIFACT_LANGUAGE = _descriptor.EnumDescriptor(
  name='Language',
  full_name='googleapis.artman.Artifact.Language',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LANGUAGE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='JAVA', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PYTHON', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NODEJS', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GO', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PHP', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CSHARP', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RUBY', index=7, number=7,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1568,
  serialized_end=1677,
)
_sym_db.RegisterEnumDescriptor(_ARTIFACT_LANGUAGE)

_ARTIFACT_ASPECT = _descriptor.EnumDescriptor(
  name='Aspect',
  full_name='googleapis.artman.Artifact.Aspect',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ALL', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CODE', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PACKAGE', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1679,
  serialized_end=1719,
)
_sym_db.RegisterEnumDescriptor(_ARTIFACT_ASPECT)


_CONFIG = _descriptor.Descriptor(
  name='Config',
  full_name='googleapis.artman.Config',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='common', full_name='googleapis.artman.Config.common', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='artifacts', full_name='googleapis.artman.Config.artifacts', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=35,
  serialized_end=136,
)


_ARTIFACT_PROTODEPENDENCY = _descriptor.Descriptor(
  name='ProtoDependency',
  full_name='googleapis.artman.Artifact.ProtoDependency',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='googleapis.artman.Artifact.ProtoDependency.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proto_path', full_name='googleapis.artman.Artifact.ProtoDependency.proto_path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=914,
  serialized_end=965,
)

_ARTIFACT_PACKAGEVERSION = _descriptor.Descriptor(
  name='PackageVersion',
  full_name='googleapis.artman.Artifact.PackageVersion',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='googleapis.artman.Artifact.PackageVersion.version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='grpc_dep_lower_bound', full_name='googleapis.artman.Artifact.PackageVersion.grpc_dep_lower_bound', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='grpc_dep_upper_bound', full_name='googleapis.artman.Artifact.PackageVersion.grpc_dep_upper_bound', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=967,
  serialized_end=1060,
)

_ARTIFACT_PUBLISHTARGET_DIRECTORYMAPPING = _descriptor.Descriptor(
  name='DirectoryMapping',
  full_name='googleapis.artman.Artifact.PublishTarget.DirectoryMapping',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='src', full_name='googleapis.artman.Artifact.PublishTarget.DirectoryMapping.src', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dest', full_name='googleapis.artman.Artifact.PublishTarget.DirectoryMapping.dest', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='googleapis.artman.Artifact.PublishTarget.DirectoryMapping.name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1262,
  serialized_end=1321,
)

_ARTIFACT_PUBLISHTARGET = _descriptor.Descriptor(
  name='PublishTarget',
  full_name='googleapis.artman.Artifact.PublishTarget',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='googleapis.artman.Artifact.PublishTarget.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='location', full_name='googleapis.artman.Artifact.PublishTarget.location', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='googleapis.artman.Artifact.PublishTarget.type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='directory_mappings', full_name='googleapis.artman.Artifact.PublishTarget.directory_mappings', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_ARTIFACT_PUBLISHTARGET_DIRECTORYMAPPING, ],
  enum_types=[
    _ARTIFACT_PUBLISHTARGET_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1063,
  serialized_end=1374,
)

_ARTIFACT = _descriptor.Descriptor(
  name='Artifact',
  full_name='googleapis.artman.Artifact',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='googleapis.artman.Artifact.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='api_name', full_name='googleapis.artman.Artifact.api_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='api_version', full_name='googleapis.artman.Artifact.api_version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='organization_name', full_name='googleapis.artman.Artifact.organization_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='release_level', full_name='googleapis.artman.Artifact.release_level', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='src_proto_paths', full_name='googleapis.artman.Artifact.src_proto_paths', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proto_deps', full_name='googleapis.artman.Artifact.proto_deps', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='test_proto_deps', full_name='googleapis.artman.Artifact.test_proto_deps', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='service_yaml', full_name='googleapis.artman.Artifact.service_yaml', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='gapic_yaml', full_name='googleapis.artman.Artifact.gapic_yaml', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='proto_package', full_name='googleapis.artman.Artifact.proto_package', index=10,
      number=18, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='samples', full_name='googleapis.artman.Artifact.samples', index=11,
      number=19, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='grpc_service_config', full_name='googleapis.artman.Artifact.grpc_service_config', index=12,
      number=20, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='googleapis.artman.Artifact.type', index=13,
      number=12, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='language', full_name='googleapis.artman.Artifact.language', index=14,
      number=13, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='package_version', full_name='googleapis.artman.Artifact.package_version', index=15,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='publish_targets', full_name='googleapis.artman.Artifact.publish_targets', index=16,
      number=15, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='discovery_doc', full_name='googleapis.artman.Artifact.discovery_doc', index=17,
      number=16, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aspect', full_name='googleapis.artman.Artifact.aspect', index=18,
      number=17, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='language_out_override', full_name='googleapis.artman.Artifact.language_out_override', index=19,
      number=21, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_ARTIFACT_PROTODEPENDENCY, _ARTIFACT_PACKAGEVERSION, _ARTIFACT_PUBLISHTARGET, ],
  enum_types=[
    _ARTIFACT_RELEASELEVEL,
    _ARTIFACT_TYPE,
    _ARTIFACT_LANGUAGE,
    _ARTIFACT_ASPECT,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=139,
  serialized_end=1744,
)

_CONFIG.fields_by_name['common'].message_type = _ARTIFACT
_CONFIG.fields_by_name['artifacts'].message_type = _ARTIFACT
_ARTIFACT_PROTODEPENDENCY.containing_type = _ARTIFACT
_ARTIFACT_PACKAGEVERSION.containing_type = _ARTIFACT
_ARTIFACT_PUBLISHTARGET_DIRECTORYMAPPING.containing_type = _ARTIFACT_PUBLISHTARGET
_ARTIFACT_PUBLISHTARGET.fields_by_name['type'].enum_type = _ARTIFACT_PUBLISHTARGET_TYPE
_ARTIFACT_PUBLISHTARGET.fields_by_name['directory_mappings'].message_type = _ARTIFACT_PUBLISHTARGET_DIRECTORYMAPPING
_ARTIFACT_PUBLISHTARGET.containing_type = _ARTIFACT
_ARTIFACT_PUBLISHTARGET_TYPE.containing_type = _ARTIFACT_PUBLISHTARGET
_ARTIFACT.fields_by_name['release_level'].enum_type = _ARTIFACT_RELEASELEVEL
_ARTIFACT.fields_by_name['proto_deps'].message_type = _ARTIFACT_PROTODEPENDENCY
_ARTIFACT.fields_by_name['test_proto_deps'].message_type = _ARTIFACT_PROTODEPENDENCY
_ARTIFACT.fields_by_name['type'].enum_type = _ARTIFACT_TYPE
_ARTIFACT.fields_by_name['language'].enum_type = _ARTIFACT_LANGUAGE
_ARTIFACT.fields_by_name['package_version'].message_type = _ARTIFACT_PACKAGEVERSION
_ARTIFACT.fields_by_name['publish_targets'].message_type = _ARTIFACT_PUBLISHTARGET
_ARTIFACT.fields_by_name['aspect'].enum_type = _ARTIFACT_ASPECT
_ARTIFACT_RELEASELEVEL.containing_type = _ARTIFACT
_ARTIFACT_TYPE.containing_type = _ARTIFACT
_ARTIFACT_LANGUAGE.containing_type = _ARTIFACT
_ARTIFACT_ASPECT.containing_type = _ARTIFACT
DESCRIPTOR.message_types_by_name['Config'] = _CONFIG
DESCRIPTOR.message_types_by_name['Artifact'] = _ARTIFACT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Config = _reflection.GeneratedProtocolMessageType('Config', (_message.Message,), dict(
  DESCRIPTOR = _CONFIG,
  __module__ = 'config_pb2'
  # @@protoc_insertion_point(class_scope:googleapis.artman.Config)
  ))
_sym_db.RegisterMessage(Config)

Artifact = _reflection.GeneratedProtocolMessageType('Artifact', (_message.Message,), dict(

  ProtoDependency = _reflection.GeneratedProtocolMessageType('ProtoDependency', (_message.Message,), dict(
    DESCRIPTOR = _ARTIFACT_PROTODEPENDENCY,
    __module__ = 'config_pb2'
    # @@protoc_insertion_point(class_scope:googleapis.artman.Artifact.ProtoDependency)
    ))
  ,

  PackageVersion = _reflection.GeneratedProtocolMessageType('PackageVersion', (_message.Message,), dict(
    DESCRIPTOR = _ARTIFACT_PACKAGEVERSION,
    __module__ = 'config_pb2'
    # @@protoc_insertion_point(class_scope:googleapis.artman.Artifact.PackageVersion)
    ))
  ,

  PublishTarget = _reflection.GeneratedProtocolMessageType('PublishTarget', (_message.Message,), dict(

    DirectoryMapping = _reflection.GeneratedProtocolMessageType('DirectoryMapping', (_message.Message,), dict(
      DESCRIPTOR = _ARTIFACT_PUBLISHTARGET_DIRECTORYMAPPING,
      __module__ = 'config_pb2'
      # @@protoc_insertion_point(class_scope:googleapis.artman.Artifact.PublishTarget.DirectoryMapping)
      ))
    ,
    DESCRIPTOR = _ARTIFACT_PUBLISHTARGET,
    __module__ = 'config_pb2'
    # @@protoc_insertion_point(class_scope:googleapis.artman.Artifact.PublishTarget)
    ))
  ,
  DESCRIPTOR = _ARTIFACT,
  __module__ = 'config_pb2'
  # @@protoc_insertion_point(class_scope:googleapis.artman.Artifact)
  ))
_sym_db.RegisterMessage(Artifact)
_sym_db.RegisterMessage(Artifact.ProtoDependency)
_sym_db.RegisterMessage(Artifact.PackageVersion)
_sym_db.RegisterMessage(Artifact.PublishTarget)
_sym_db.RegisterMessage(Artifact.PublishTarget.DirectoryMapping)


# @@protoc_insertion_point(module_scope)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import mock
import os
import unittest

from artman.tasks import python_grpc_tasks


class PythonPackageChangeTest(unittest.TestCase):

    _TASK = python_grpc_tasks.PythonChangePackageTask()
    _TASK._organization_name = 'google'

    _PROTO_FILE = [
        '# Comment line\n',
        'package google.service.v1;\n',
        'import "google/service/v1/a.proto";\n',
        'import "google/cloud/otherapi/v3/b.proto";\n',
        'import "google/common/common_proto.proto";\n',
        'Some other text referencing to google.service.v1\n']

    def test__extract_base_dirs(self):
        mock_proto = mock.mock_open()
        mock_proto.return_value.__iter__ = lambda _: iter(self._PROTO_FILE)
        with mock.patch.object(io, 'open', mock_proto):
            base_dirs = self._TASK._extract_base_dirs(
                os.path.join('a', 'test', 'path', 'to', 'google', 'service',
                             'v1', 'a.proto'))
            repeated_base_dirs = self._TASK._extract_base_dirs(
                os.path.join('home', 'foo', 'workspaces', 'google', 'some',
                             'v1', 'googleapis', 'google', 'service', 'v1',
                             'a.proto'))
        expected = os.path.join('google', 'service', 'v1')
        self.assertEqual(base_dirs, expected)
        self.assertEqual(repeated_base_dirs, expected)

    def test__transfom(self):
        # Simple package transformations with arbitrary separator
        self.assertEqual(self._TASK._transform('google.service.v1', '.', []),
                         'google.service_v1.proto')
        self.assertEqual(
            self._TASK._transform('google.service.v1alpha', '.', []),
            'google.service_v1alpha.proto',
        )
        self.assertEqual(self._TASK._transform('google/other/v1', '/', []),
                         'google/other_v1/proto')
        self.assertEqual(self._TASK._transform('google$service', '$', []),
                         'google$service')

        # Don't transform common protos
        self.assertEqual(
            self._TASK._transform('google/common', '/', ['google.common']),
            'google/common')
        self.assertEqual(
            self._TASK._transform(
                'google/uncommon/v1',
                '/',
                ['google.common'],
            ),
            'google/uncommon_v1/proto',
        )

        # Don't transform non-Google protos
        self.assertEqual(
            self._TASK._transform('my_custom/path', '/', ['']),
            'my_custom/path')

    def test__copy_proto(self):
        mock_proto = mock.mock_open()
        mock_proto.return_value.__iter__ = lambda _: iter(self._PROTO_FILE)
        with mock.patch.object(io, 'open', mock_proto):
            self._TASK._copy_proto('foo', 'bar', ['google.common'])
        expected_writes = [
            mock.call('# Comment line\n'),
            mock.call('package google.service.v1;\n'),
            mock.call('import "google/service_v1/proto/a.proto";\n'),
            mock.call('import "google/cloud/otherapi_v3/proto/b.proto";\n'),
            mock.call('import "google/common/common_proto.proto";\n'),
            mock.call('Some other text referencing to google.service.v1\n')]

        mock_proto().write.assert_has_calls(expected_writes)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tasks related to protoc"""

import io
import json
import os
import re
from ruamel import yaml

import six

from artman.tasks import task_base
from artman.utils import task_utils
from artman.utils.logger import logger
from artman.utils import protoc_utils


class ProtoDescGenTask(task_base.TaskBase):
    """Generates proto descriptor set"""
    default_provides = 'descriptor_set'

    def execute(self, src_proto_path, import_proto_path, output_dir,
                api_name, api_version, organization_name, toolkit_path,
                root_dir, excluded_proto_path=[], proto_deps=[], language='python'):
        desc_proto_paths = []
        for dep in proto_deps:
            if 'proto_path' in dep and dep['proto_path']:
                desc_proto_paths.append(os.path.join(root_dir, dep['proto_path']))
        desc_protos = list(
            protoc_utils.find_protos(src_proto_path + desc_proto_paths,
                                     excluded_proto_path))
        header_proto_path = import_proto_path + desc_proto_paths
        header_proto_path.extend(src_proto_path)
        desc_out_file = task_utils.api_full_name(
            api_name, api_version, organization_name) + '.desc'
        logger.debug('Compiling descriptors for {0}'.format(desc_protos))
        self.exec_command(['mkdir', '-p', output_dir])

        proto_params = protoc_utils.PROTO_PARAMS_MAP[language]

        proto_compiler_command = proto_params.proto_compiler_command
        logger.debug('Using protoc command: {0}'.format(proto_compiler_command))
        # DescGen doesn't use group protos by package right now because
        #   - it doesn't have to
        #   - and multiple invocation will overwrite the desc_out_file
        (common_resources_includes, common_resources_paths) = \
            protoc_utils.protoc_common_resources_params(root_dir)
        params = proto_params.proto_compiler_command + \
            common_resources_includes + \
            protoc_utils.protoc_header_params(header_proto_path, toolkit_path) + \
            protoc_utils.protoc_desc_params(output_dir, desc_out_file) + \
            common_resources_paths + \
            desc_protos

        self.exec_command(params)
        return os.path.join(output_dir, desc_out_file)


class ProtocCodeGenTaskBase(task_base.TaskBase):
    """Generates protos"""
    def _execute_proto_codegen(
            self, language, src_proto_path, import_proto_path,
            pkg_dir, api_name, api_version, organization_name,
            toolkit_path, gapic_yaml, root_dir,
            gen_proto=False, gen_grpc=False, gen_common_resources=False,
            final_src_proto_path=None, final_import_proto_path=None,
            excluded_proto_path=[], language_out_override=None):
        # Adding 17th parameter is a sin that I commit here just because
        # refactoring of this code will never happen.
        src_proto_path = final_src_proto_path or src_proto_path
        import_proto_path = final_import_proto_path or import_proto_path
        proto_params = protoc_utils.PROTO_PARAMS_MAP[language]

        if gen_proto:
            protoc_proto_params = protoc_utils.protoc_proto_params(
                proto_params, pkg_dir, gapic_yaml, with_grpc=True, 
                language_out_override=language_out_override)
        else:
            protoc_proto_params = []

        if gen_grpc:
            protoc_grpc_params = protoc_utils.protoc_grpc_params(
                proto_params, pkg_dir, toolkit_path)
        else:
            protoc_grpc_params = []

        if gen_common_resources:
            (common_resources_includes, common_resources_paths) = \
                protoc_utils.protoc_common_resources_params(root_dir)
            protoc_plugin_params = protoc_utils.protoc_plugin_params(
                proto_params,  pkg_dir, gapic_yaml)
        else:
            (common_resources_includes, common_resources_paths) = ([], [])
            protoc_plugin_params = []

        if not protoc_proto_params \
                and not protoc_grpc_params \
                and not protoc_plugin_params:
            return pkg_dir

        # protoc-gen-go has some peculiarities:
        # It can only compile one package per invocation. So, we need to split
        # proto files by packages.
        #
        # The order of the input files affects comments and internal variables.
        # While this doesn't affect the correctness of the result, we sort
        # proto files for reproducibility.

        # For other languages, we'll pass all proto files into the same protoc
        # invocation (PHP especially needs that).
        all_protos = protoc_utils.find_protos(src_proto_path, excluded_proto_path)
        if language == "go":
            protos_map = protoc_utils.group_by_go_package(all_protos)
        else:
            protos_map = { "": all_protos }

        for (dirname, protos) in protos_map.items():
            # It is possible to get duplicate protos. De-dupe them.
            protos = sorted(set(protos))

            command_params = proto_params.proto_compiler_command + \
                common_resources_includes + \
                protoc_utils.protoc_header_params(
                    import_proto_path + src_proto_path, toolkit_path) + \
                protoc_proto_params + \
                protoc_grpc_params + \
                protoc_plugin_params + \
                common_resources_paths + \
                protos

            # Execute protoc.
            self.exec_command(command_params)

        return pkg_dir


class ProtoCodeGenTask(ProtocCodeGenTaskBase):
    default_provides = 'proto_code_dir'

    """Generates protos"""
    def execute(self, language, src_proto_path, import_proto_path,
                output_dir, api_name, api_version, organization_name,
                toolkit_path, gapic_yaml, root_dir, final_src_proto_path=None,
                final_import_proto_path=None, excluded_proto_path=[],
                language_out_override=None):
        pkg_dir = protoc_utils.prepare_proto_pkg_dir(
            output_dir, api_name, api_version, organization_name, language)
        return self._execute_proto_codegen(
            language, src_proto_path, import_proto_path, pkg_dir,
            api_name, api_version, organization_name, toolkit_path,
            gapic_yaml, root_dir, gen_proto=True,
            final_src_proto_path=final_src_proto_path,
            final_import_proto_path=final_import_proto_path,
            excluded_proto_path=excluded_proto_path,
            language_out_override=language_out_override)

class ResourceNameGenTask(ProtocCodeGenTaskBase):
    default_provides = 'proto_code_dir'

    """Generates protos"""
    def execute(self, language, src_proto_path, import_proto_path,
                output_dir, api_name, api_version, organization_name,
                toolkit_path, gapic_yaml, root_dir, final_src_proto_path=None,
                final_import_proto_path=None, excluded_proto_path=[],
                language_out_override=None):
        pkg_dir = protoc_utils.prepare_proto_pkg_dir(
            output_dir, api_name, api_version, organization_name, language)
        return self._execute_proto_codegen(
            language, src_proto_path, import_proto_path, pkg_dir,
            api_name, api_version, organization_name, toolkit_path,
            gapic_yaml, root_dir, gen_common_resources=True,
            final_src_proto_path=final_src_proto_path,
            final_import_proto_path=final_import_proto_path,
            excluded_proto_path=excluded_proto_path,
            language_out_override=language_out_override)

class GrpcCodeGenTask(ProtocCodeGenTaskBase):
    default_provides = 'grpc_code_dir'

    """Generates the gRPC client library"""
    def execute(self, language, src_proto_path, import_proto_path,
                toolkit_path, output_dir, api_name, api_version,
                organization_name, gapic_yaml, root_dir, final_src_proto_path=None,
                final_import_proto_path=None, excluded_proto_path=[],
                language_out_override=None):
        pkg_dir = protoc_utils.prepare_grpc_pkg_dir(
            output_dir, api_name, api_version, organization_name, language)
        return self._execute_proto_codegen(
            language, src_proto_path, import_proto_path, pkg_dir,
            api_name, api_version,  organization_name, toolkit_path,
            gapic_yaml, root_dir, gen_grpc=True,
            final_src_proto_path=final_src_proto_path,
            final_import_proto_path=final_import_proto_path,
            excluded_proto_path=excluded_proto_path,
            language_out_override=language_out_override)


class ProtoAndGrpcCodeGenTask(ProtocCodeGenTaskBase):
    default_provides = 'grpc_code_dir'

    """Generates protos and the gRPC client library"""
    def execute(self, language, src_proto_path, import_proto_path,
                toolkit_path, output_dir, api_name, api_version,
                organization_name, gapic_yaml, root_dir, final_src_proto_path=None,
                final_import_proto_path=None, excluded_proto_path=[],
                language_out_override=None):
        pkg_dir = protoc_utils.prepare_grpc_pkg_dir(
            output_dir, api_name, api_version, organization_name, language)
        return self._execute_proto_codegen(
            language, src_proto_path, import_proto_path, pkg_dir,
            api_name, api_version, organization_name, toolkit_path,
            gapic_yaml, root_dir, gen_proto=True, gen_grpc=True,
            final_src_proto_path=final_src_proto_path,
            final_import_proto_path=final_import_proto_path,
            excluded_proto_path=excluded_proto_path,
            language_out_override=language_out_override)


class GoCopyTask(task_base.TaskBase):
    def execute(self, gapic_code_dir, grpc_code_dir):
        for entry in os.listdir(grpc_code_dir):
            src_path = os.path.join(grpc_code_dir, entry)
            self.exec_command([
                'cp', '-rf', src_path, gapic_code_dir])


class RubyGrpcCopyTask(task_base.TaskBase):
    """Copies the generated protos and gRPC client library to
    the gapic_code_dir/lib.
    """
    def execute(self, api_name, api_version, language, organization_name,
                output_dir, gapic_code_dir, grpc_code_dir):
        final_output_dir = os.path.join(gapic_code_dir, 'lib')
        logger.info('Copying %s/* to %s.' % (grpc_code_dir, final_output_dir))
        if not os.path.exists(final_output_dir):
            self.exec_command(['mkdir', '-p', final_output_dir])
        for entry in sorted(os.listdir(grpc_code_dir)):
            src_path = os.path.join(grpc_code_dir, entry)
            self.exec_command([
                'cp', '-rf', src_path, final_output_dir])


class JavaProtoCopyTask(task_base.TaskBase):
    """Copies the .proto files into the grpc_code_dir directory
    """
    def execute(self, src_proto_path, proto_code_dir, excluded_proto_path=[]):
        grpc_proto_dir = os.path.join(proto_code_dir, 'src', 'main', 'proto')
        for proto_path in src_proto_path:
            index = protoc_utils.find_google_dir_index(proto_path)
            for src_proto_file in protoc_utils.find_protos(
                    [proto_path], excluded_proto_path):
                relative_proto_file = src_proto_file[index:]
                dst_proto_file = os.path.join(
                    grpc_proto_dir, relative_proto_file)
                self.exec_command(
                    ['mkdir', '-p', os.path.dirname(dst_proto_file)])
                self.exec_command(['cp', src_proto_file, dst_proto_file])


class PhpGrpcMoveTask(task_base.TaskBase):
    """Moves the generated protos and gRPC client library to
    the gapic_code_dir/proto directory.
    """
    default_provides = 'grpc_code_dir'

    def execute(self, grpc_code_dir, gapic_code_dir=None):
        if not gapic_code_dir:
            return grpc_code_dir
        final_output_dir = os.path.join(gapic_code_dir, 'proto')
        if not os.path.exists(final_output_dir):
            self.exec_command(['mkdir', '-p', final_output_dir])
        logger.info('Moving %s/* to %s.' % (grpc_code_dir, final_output_dir))
        for entry in sorted(os.listdir(grpc_code_dir)):
            src_path = os.path.join(grpc_code_dir, entry)
            self.exec_command([
                'mv', src_path, os.path.join(final_output_dir, entry)])
        self.exec_command([
            'rm', '-r', grpc_code_dir])
        return final_output_dir


# TODO (michaelbausor): Once correct naming is supported in
# gRPC, we should remove this.
class PhpGrpcRenameTask(task_base.TaskBase):
    """Rename references to proto files in the gRPC stub."""

    def execute(self, grpc_code_dir):
        for filename in protoc_utils.list_files_recursive(grpc_code_dir):
            if filename.endswith('GrpcClient.php'):
                logger.info('Performing replacements in: %s' % (filename,))
                with io.open(filename, encoding='UTF-8') as f:
                    contents = f.read()
                contents = protoc_utils.php_proto_rename(contents)
                with io.open(filename, 'w', encoding='UTF-8') as f:
                    f.write(contents)


class NodeJsProtoCopyTask(task_base.TaskBase):
    """Copies the .proto files into the gapic_code_dir/proto directory
    and compiles these proto files to protobufjs JSON.
    """
    def execute(self, gapic_code_dir, src_proto_path, excluded_proto_path=[]):
        final_output_dir = os.path.join(gapic_code_dir, 'protos')
        src_dir = os.path.join(gapic_code_dir, 'src')
        proto_files = []
        for proto_path in src_proto_path:
            index = protoc_utils.find_google_dir_index(proto_path)
            for src_proto_file in protoc_utils.find_protos(
                    [proto_path], excluded_proto_path):
                relative_proto_file = src_proto_file[index:]
                proto_files.append(relative_proto_file)
                dst_proto_file = os.path.join(
                    final_output_dir, relative_proto_file)
                dst_proto_dir = os.path.dirname(dst_proto_file)
                if not os.path.exists(dst_proto_dir):
                    self.exec_command(['mkdir', '-p', dst_proto_dir])
                self.exec_command(['cp', src_proto_file, dst_proto_file])
        # Execute compileProtos from Docker image (a part of from google-gax)
        cwd = os.getcwd()
        os.chdir(gapic_code_dir)
        self.exec_command(['compileProtos', './src'])
        os.chdir(cwd)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pipelines that run gRPC codegen"""

from artman.pipelines import code_generation as code_gen
from artman.tasks import emit_success
from artman.tasks import package_metadata_tasks
from artman.tasks import protoc_tasks
from artman.tasks import python_grpc_tasks
from artman.utils import task_utils


class GrpcClientPipeline(code_gen.CodeGenerationPipelineBase):

    def __init__(self, **kwargs):
        super(GrpcClientPipeline, self).__init__(
            ProtoGenTaskFactory(gen_grpc=True, **kwargs), **kwargs)


class ProtoClientPipeline(code_gen.CodeGenerationPipelineBase):

    def __init__(self, **kwargs):
        super(ProtoClientPipeline, self).__init__(
            ProtoGenTaskFactory(gen_grpc=False, **kwargs), **kwargs)


class ProtoGenTaskFactory(code_gen.TaskFactoryBase):

    def __init__(self, gen_grpc, **kwargs):
        if 'language' not in kwargs:
            raise ValueError('Valid `language` argument required for gRPC codegen')
        if 'aspect' not in kwargs:
            raise ValueError('Valid `aspect` argument required for gRPC codegen')
        self.language = kwargs['language']
        self.gen_grpc = gen_grpc
        self.gen_code = kwargs['aspect'] == 'ALL' or kwargs['aspect'] == 'CODE'
        self.gen_pkg = kwargs['aspect'] == 'ALL' or kwargs['aspect'] == 'PACKAGE'

    def get_tasks(self, **kwargs):
        tasks = self.get_grpc_codegen_tasks(**kwargs)
        tasks += emit_success.TASKS
        return task_utils.instantiate_tasks(tasks, kwargs)

    def get_grpc_codegen_tasks(self, **kwargs):
        methods = {
            'java': self._get_grpc_codegen_tasks_java,
            'python': self._get_grpc_codegen_tasks_python,
            'go': self._get_grpc_codegen_tasks_go,
            'ruby': self._get_grpc_codegen_tasks_ruby,
            'php': self._get_grpc_codegen_tasks_php,
            'csharp': self._get_grpc_codegen_tasks_csharp,
            'nodejs': self._get_grpc_codegen_tasks_nodejs
        }
        return methods[self.language](**kwargs)

    def _get_grpc_codegen_tasks_java(self, **kwargs):
        tasks = [protoc_tasks.ProtoDescGenTask]
        if self.gen_code:
            tasks.append(protoc_tasks.ProtoCodeGenTask)
            tasks.append(protoc_tasks.ResourceNameGenTask)
            if self.gen_grpc:
                tasks.append(protoc_tasks.GrpcCodeGenTask)
        tasks.append(package_metadata_tasks.PackageMetadataConfigGenTask)
        if self.gen_pkg:
            tasks.append(package_metadata_tasks.ProtoPackageMetadataGenTask)
            if self.gen_grpc:
                tasks.append(package_metadata_tasks.GrpcPackageMetadataGenTask)
        tasks.append(protoc_tasks.JavaProtoCopyTask)
        return tasks

    def _get_grpc_codegen_tasks_python(self, **kwargs):
        tasks = [python_grpc_tasks.PythonChangePackageTask]
        tasks.append(protoc_tasks.ProtoDescGenTask)
        if self.gen_code:
            tasks.append(protoc_tasks.ProtoAndGrpcCodeGenTask)
        if self.gen_pkg:
            tasks.append(package_metadata_tasks.PackageMetadataConfigGenTask)
        tasks.append(python_grpc_tasks.PythonMoveProtosTask)
        return tasks

    def _get_grpc_codegen_tasks_go(self, **kwargs):
        return [
            protoc_tasks.ProtoAndGrpcCodeGenTask,
            protoc_tasks.GoCopyTask,
        ]

    def _get_grpc_codegen_tasks_ruby(self, **kwargs):
        return [
            protoc_tasks.ProtoAndGrpcCodeGenTask,
            protoc_tasks.RubyGrpcCopyTask,
        ]

    def _get_grpc_codegen_tasks_php(self, **kwargs):
        return [
            protoc_tasks.ProtoAndGrpcCodeGenTask,
            protoc_tasks.PhpGrpcRenameTask,
            protoc_tasks.PhpGrpcMoveTask,
        ]
    def _get_grpc_codegen_tasks_csharp(self, **kwargs):
        return [protoc_tasks.ProtoCodeGenTask, protoc_tasks.GrpcCodeGenTask]

    def _get_grpc_codegen_tasks_nodejs(self, **kwargs):
        return [protoc_tasks.NodeJsProtoCopyTask]

    def get_validate_kwargs(self):
        if self.language == 'go':
            return ['gapic_yaml', 'gapic_code_dir'] + code_gen.COMMON_REQUIRED
        else:
            return code_gen.COMMON_REQUIRED

    def get_invalid_kwargs(self):
        return []



# Copyright 2017 Google
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from colorlog import ColoredFormatter

__all__ = ('logger', 'output_logger', 'setup_logging')


OUTPUT = 22
SUCCESS = 25

COLORS = {
    'DEBUG': 'blue',
    'INFO': 'cyan',
    'OUTPUT': 'white',
    'SUCCESS': 'green',
    'WARNING': 'yellow',
    'ERROR': 'red',
    'CRITICAL': 'red,bold',
}


class Logger(logging.getLoggerClass()):
    def __init__(self, name, level=logging.NOTSET):
        super(Logger, self).__init__(name, level)
        logging.addLevelName(OUTPUT, 'OUTPUT')
        logging.addLevelName(SUCCESS, 'SUCCESS')

    def output(self, message, *args, **kwargs):
        if self.isEnabledFor(OUTPUT):
            self._log(OUTPUT, message, args, **kwargs)

    def success(self, message, *args, **kwargs):
        if self.isEnabledFor(SUCCESS):
            self._log(SUCCESS, message, args, **kwargs)


logging.setLoggerClass(Logger)


def setup_logging(level=logging.DEBUG):  # pragma: no cover
    setup_logger(None, level)
    setup_logger('artman.output', level + 5,
        colors=dict(COLORS, OUTPUT='green'),
        format_string="%(log_color)s%(name)s >\n%(white)s%(message)s",
    )
    setup_logger('github3', level + 10, colors=dict(COLORS, INFO='blue'))
    setup_logger('sh', logging.WARNING)


def setup_logger(name, level, colors=COLORS,
                 format_string="%(purple)s%(name)s> %(log_color)s%(message)s"):
    """Set up a particular logger with overridden behavior.

    Args:
        name (str): The name of the logger.
        level (int): The log level to set this logger to.
        colors (dict): A dictionary of log colors.
        format_string (str): The format of the log message.
    """
    logger_ = logging.getLogger(name)
    logger_.setLevel(level)
    formatter = ColoredFormatter(format_string, reset=True, log_colors=colors)
    if logger_.handlers:
        logger_.handlers.pop()
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger_.addHandler(handler)
    logger_.propagate = False
    return logger_


# Make the logger and output_logger available for import from elsewhere.
logger = logging.getLogger('artman')
output_logger = logging.getLogger('artman.output')



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tasks related to format"""

import os
import subprocess

from artman.tasks import task_base
from artman.utils import task_utils
from artman.utils.logger import logger


# TODO: Store both intermediate and final output in all format tasks.

class JavaFormatTask(task_base.TaskBase):
    def execute(self, gapic_code_dir, toolkit_path):
        logger.debug('Formatting files in %s.' %
                    os.path.abspath(gapic_code_dir))
        path = task_utils.get_java_tool_path(toolkit_path, 'googleJavaFormatJar')
        targetFiles = []
        for root, dirs, files in os.walk(gapic_code_dir):
            for filename in files:
                if filename.endswith('.java'):
                    targetFile = os.path.abspath(os.path.join(root, filename))
                    targetFiles.append(targetFile)
        self.exec_command(
                ['java', '-jar', path, '--replace'] + targetFiles)

    def validate(self):
        return []


class GoFormatTask(task_base.TaskBase):
    def execute(self, gapic_code_dir):
        logger.debug('Formatting files in %s.' %
                    os.path.abspath(gapic_code_dir))
        self.exec_command(['gofmt', '-w', gapic_code_dir])


class PhpFormatTask(task_base.TaskBase):
    def execute(self, gapic_code_dir):
        abs_code_dir = os.path.abspath(gapic_code_dir)
        logger.debug('Formatting file using php-cs-fixer in %s.' % abs_code_dir)
        subprocess.call(['php-cs-fixer', 'fix',
                         '--rules=@Symfony,-phpdoc_annotation_without_dot',
                         gapic_code_dir])
        # We require a second call to php-cs-fixer because instances of @type
        # have been converted to @var. We cannot disable this conversion in
        # the first call without affecting other aspects of the formatting.
        subprocess.call(['php-cs-fixer',
                         'fix',
                         '--rules={"phpdoc_no_alias_tag" : {"replacements" : '
                         '{"var" : "type"}}}',
                         gapic_code_dir])
        logger.debug('Formatting file using phpcbf in %s.' % abs_code_dir)
        subprocess.call(['phpcbf', '--standard=PSR2', '--no-patch',
                         gapic_code_dir])


_FORMAT_TASK_DICT = {
    'java': JavaFormatTask,
    'go': GoFormatTask,
    'php': PhpFormatTask,
}


def get_format_task(language):
    return _FORMAT_TASK_DICT.get(language, task_base.EmptyTask)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils related to config files"""

from __future__ import absolute_import
import io

from ruamel import yaml

import six


def load_config_spec(all_config_data, language):
    segments = [all_config_data['common']]
    segments.append(all_config_data.get(language, {}))
    config = merge(*segments)
    return config

def merge(*dictionaries):
    """Recursively merge one or more dictionaries together.

    Conflict resolution:
        - Dictionary values are recursively merged.
        - List values are added together.
        - Sets are combined with `__or__` (equivalent of set.union).
        - Everything else, the right-most value wins.
        - If values are type mismatches in the above resolution rules,
          raise ValueError.

    Args:
        *dictionaries (list(dict)): Dictionaries to be merged.

    Returns:
        dict: The merged dictionary.
    """
    answer = {}
    for d in dictionaries:
        for k, v in six.iteritems(d):
            # If the value is not yet found in the answer, then
            # simply add it and move on.
            if k not in answer:
                answer[k] = v
                continue

            # If the values are both lists, append them to one another.
            if isinstance(answer[k], list):
                if not isinstance(v, list):
                    raise ValueError('Attempt to merge a list and non-list.')
                answer[k] = answer[k] + v
                continue

            # If the values are both sets, take their union.
            if isinstance(answer[k], set):
                if not isinstance(v, set):
                    raise ValueError('Attempt to merge a set and non-set.')
                answer[k] = answer[k] | (v)
                continue

            # If the values are both dictionaries, merge them recursively
            # by calling this method.
            if isinstance(answer[k], dict):
                if not isinstance(v, dict):
                    raise ValueError('Attempt to merge a dict and non-dict.')
                answer[k] = merge(answer[k], v)
                continue

            # These are either primitives or other objects; the right-hand
            # value wins.
            answer[k] = v
    return answer



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Artman config converter.

It converts the new artman config file into the legacy format so that current
artman code can work with both legacy and new artman config file. We will phase
out legacy artman config yamls and make artman code read the configuration from
the new artman config. Once that is done, this converter can be removed.
"""

from __future__ import absolute_import
import os
from protobuf_to_dict import protobuf_to_dict

from artman.config.proto.config_pb2 import Artifact
from artman.utils.logger import logger


def convert_to_legacy_config_dict(artifact_config, root_dir, output_dir):
    artifact_config_dict = protobuf_to_dict(artifact_config)
    common = {}
    common['api_name'] = artifact_config.api_name
    common['api_version'] = artifact_config.api_version
    common['organization_name'] = artifact_config.organization_name
    common['service_yaml'] = artifact_config.service_yaml
    common['gapic_yaml'] = artifact_config.gapic_yaml
    common['proto_package'] = artifact_config.proto_package
    if 'grpc_service_config' in artifact_config_dict:
        common['grpc_service_config'] = artifact_config_dict['grpc_service_config']
    common['samples'] = artifact_config.samples
    src_proto_paths = artifact_config_dict.get('src_proto_paths', [])
    common['src_proto_path'], excluded_proto_path = _calculate_proto_paths(src_proto_paths)
    if excluded_proto_path:
        common['excluded_proto_path'] = excluded_proto_path
    common['import_proto_path'] = [root_dir]
    common['output_dir'] = output_dir
    common['proto_deps'] = []
    if 'proto_deps' in artifact_config_dict:
        common['proto_deps'] = artifact_config_dict['proto_deps']
    if 'test_proto_deps' in artifact_config_dict:
        common['test_proto_deps'] = artifact_config_dict['test_proto_deps']
    common['artifact_type'] = Artifact.Type.Name(artifact_config.type)
    common['language_out_override'] = artifact_config.language_out_override

    result = {}
    result['common'] = common

    if artifact_config.type == Artifact.GAPIC_CONFIG\
        or artifact_config.type == Artifact.DISCOGAPIC_CONFIG:
        # Early return if the artifact type is GAPIC_CONFIG or DISCOGAPIC_CONFIG
        return result

    language = Artifact.Language.Name(
        artifact_config.language).lower()
    language_config_dict = {}
    rel_gapic_code_dir = _calculate_rel_gapic_output_dir(
        language, artifact_config.api_name, artifact_config.api_version)
    language_config_dict['gapic_code_dir'] = os.path.join(
        output_dir, rel_gapic_code_dir)
    if artifact_config.release_level != Artifact.RELEASE_LEVEL_UNSPECIFIED:
        language_config_dict['release_level'] = (
            Artifact.ReleaseLevel.Name(
                artifact_config.release_level).lower())

    result[language] = language_config_dict
    return result


def _calculate_rel_gapic_output_dir(language, api_name, api_version):
    """Calculate the gapic output dir relative to the specified output_dir.

    TODO(ethanbao): This part can become part of normalization step when gapic
    ouput dir becomes configurable. This logic doesn't work for non-cloud.
    """
    if language == 'java':
        return 'java/gapic-google-cloud-%s-%s' % (api_name, api_version)
    elif language == 'csharp':
        return 'csharp/google-cloud-%s' % api_name
    elif language == 'go':
        return 'gapi-cloud-%s-go' % api_name
    elif language == 'nodejs':
        return 'js/%s-%s' % (api_name, api_version)
    elif language == 'php':
        return 'php/google-cloud-%s-%s' % (api_name, api_version)
    elif language == 'python':
        return 'python/%s-%s' % (api_name, api_version)
    elif language == 'ruby':
        api_name = api_name.replace('-', '_')
        return 'ruby/google-cloud-ruby/google-cloud-%s' % api_name

    raise ValueError('Language `%s` is not currently supported.' % language)

def _calculate_proto_paths(proto_paths):
    src_proto_path, excluded_proto_path = [], []
    for proto_path in proto_paths:
        if proto_path.startswith('-'):
            excluded_proto_path.append(proto_path[1:])
        else:
            src_proto_path.append(proto_path)
    return src_proto_path, excluded_proto_path



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tasks related to package metadata"""

import io
import os

from ruamel import yaml

from artman.tasks import task_base
from artman.utils import task_utils

# Metadata config gen

class PackageMetadataConfigGenTask(task_base.TaskBase):
    """Generates package metadata config"""
    default_provides = 'package_metadata_yaml'

    def execute(self, api_name, api_version, organization_name, output_dir,
                proto_deps, language, root_dir, src_proto_path,
                artifact_type, release_level=None,
                test_proto_deps=None):
        api_full_name = task_utils.api_full_name(
            api_name, api_version, organization_name)

        config = self._create_config(
            api_name, api_version, organization_name, output_dir, proto_deps,
            test_proto_deps, language, root_dir, src_proto_path,
            artifact_type, release_level)

        package_metadata_config = os.path.join(
            output_dir, language + '_' + api_full_name + '_package2.yaml')
        self._write_yaml(config, package_metadata_config)

        return package_metadata_config

    def _create_config(self, api_name, api_version, organization_name, output_dir,
                       proto_deps, test_proto_deps, language, root_dir, src_proto_path,
                       artifact_type, release_level):
        googleapis_dir = root_dir
        googleapis_path = os.path.commonprefix(
            [os.path.relpath(p, googleapis_dir) for p in src_proto_path])

        config = {
            'api_name': api_name,
            'api_version': api_version,
            'organization_name': organization_name,
            'proto_deps': proto_deps,
            'release_level': release_level,
            'artifact_type': artifact_type,
            'proto_path': googleapis_path,
        }

        if test_proto_deps:
            config['test_proto_deps'] = test_proto_deps

        return config

    # Separated so that this can be mocked for testing
    def _write_yaml(self, config_dict, dest):
        with io.open(dest, 'w', encoding='UTF-8') as f:
            yaml.safe_dump(config_dict, f, default_flow_style=False)

# Metadata gen

class ProtoPackageMetadataGenTaskBase(task_base.TaskBase):
    def _execute(self, api_name, api_version, organization_name, toolkit_path,
                descriptor_set, src_proto_path, service_yaml,
                input_dir, output_dir, package_metadata_yaml,
                language, artifact_type):
        api_full_name = task_utils.api_full_name(
            api_name, api_version, organization_name)
        proto_prefix = self._get_proto_prefix()

        pkg_dir = os.path.join(
            output_dir, language, proto_prefix + api_full_name)

        args = [
            '--descriptor_set=' + os.path.abspath(descriptor_set),
            '--input=' + os.path.abspath(input_dir),
            '--output=' + os.path.abspath(pkg_dir),
            '--package_yaml2=' + os.path.abspath(package_metadata_yaml),
            '--artifact_type=' + artifact_type,
            '--language=' + language,
        ]
        if service_yaml:
            args = args + ['--service_yaml=' + os.path.abspath(service_yaml)]
        self.exec_command(
            task_utils.gapic_gen_task(toolkit_path, ['LEGACY_GRPC_PACKAGE'] + args))

        return pkg_dir

    # Separated so that this can be overriden by each language subclass
    def _get_proto_prefix(self):
        return 'proto-'

class ProtoPackageMetadataGenTask(ProtoPackageMetadataGenTaskBase):
    default_provides = 'proto_pkg_dir'

    def execute(self, api_name, api_version, organization_name, toolkit_path,
                descriptor_set, src_proto_path, service_yaml,
                proto_code_dir, output_dir, package_metadata_yaml,
                language):
        return self._execute(api_name, api_version, organization_name,
                toolkit_path, descriptor_set, src_proto_path, service_yaml,
                proto_code_dir, output_dir, package_metadata_yaml,
                language, 'PROTOBUF')

class GrpcPackageMetadataGenTask(ProtoPackageMetadataGenTaskBase):
    default_provides = 'grpc_pkg_dir'

    def execute(self, api_name, api_version, organization_name, toolkit_path,
                descriptor_set, src_proto_path, service_yaml,
                grpc_code_dir, output_dir, package_metadata_yaml,
                language):
        return self._execute(api_name, api_version, organization_name,
                toolkit_path, descriptor_set, src_proto_path, service_yaml,
                grpc_code_dir, output_dir, package_metadata_yaml,
                language, 'GRPC')

    def _get_proto_prefix(self):
        return 'grpc-'



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module containing all artman tasks.

This module provides all tasks from its submodules.
It drops all "_tasks" suffices from the submodule names, which will allow
us to move to a clearer and more succinct spelling: `tasks.{domain}.{name}`
"""

from __future__ import absolute_import

from artman.tasks import emit_success as success
from artman.tasks import format_tasks as format
from artman.tasks import gapic_tasks as gapic
from artman.tasks import descriptor_set_tasks as descriptor
from artman.tasks import io_tasks as io
from artman.tasks import package_metadata_tasks as package_metadata
from artman.tasks import protoc_tasks as protoc
from artman.tasks import python_grpc_tasks as python_grpc
from artman.tasks.task_base import Task, EmptyTask

__all__ = (
    'EmptyTask', 'format', 'gapic', 'io',
    'package_metadata', 'protoc', 'python_grpc', 'success'
    'Task',
)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tasks related to Python gRPC code generation"""

import io
import os
import re
import tempfile
import time

from ruamel import yaml

from artman.utils import protoc_utils
from artman.tasks import task_base


class PythonChangePackageTask(task_base.TaskBase):
    """Copies source protos to a package that meets Python convention"""
    default_provides = ('final_src_proto_path',
                        'final_import_proto_path')

    _IDENTIFIER = '[A-Za-z_][A-Za-z_0-9]*'

    _BASE_PROTO_REGEX = (
        '(?P<prefix>{prefix})' +
        '(?P<package>' + _IDENTIFIER +
        '({separator}' + _IDENTIFIER + ')*{package_suffix})'
        '(?P<suffix>{suffix})')

    # E.g., `package google.foo.bar`
    _PACKAGE_REGEX = re.compile(_BASE_PROTO_REGEX.format(
        prefix='^package ',
        separator='\\.',
        package_suffix='',
        suffix=''))

    # E.g., `import "google/foo/bar";`
    _IMPORT_REGEX = re.compile(_BASE_PROTO_REGEX.format(
        prefix='^import (?:public )?"',
        separator='/',
        package_suffix='\\.proto',
        suffix='";'))

    # TODO (geigerj): add regex for documentation link updates?

    def execute(self, src_proto_path, import_proto_path,
                organization_name):
        self._organization_name = organization_name

        # Treat google.protobuf, google.iam as a common proto package, even
        # though they are not included in the common-protos we generate.
        #
        # TODO (geigerj): remove 'google.iam' when it is included in the common
        # protos package.
        common_protos = [
            'google.protobuf',
            'google.iam',
            'google.api',
            'google.longrunning',
            'google.rpc',
            'google.type',
            'google.logging.type',
        ]

        tmpdir = os.path.join(
            tempfile.gettempdir(), 'artman-python', str(int(time.time())))
        new_proto_dir = os.path.join(tmpdir, 'proto')
        new_src_path = set()
        new_import_path = [new_proto_dir]

        self._copy_and_transform_directories(
            src_proto_path, new_proto_dir, common_protos, paths=new_src_path)
        self._copy_and_transform_directories(
            import_proto_path, new_proto_dir, common_protos)

        # Update src_proto_path, import_proto_path
        return list(new_src_path), new_import_path

    def _extract_base_dirs(self, proto_file):
        """Return the proto file path derived from the package name."""

        with io.open(proto_file, 'rt', encoding='UTF-8') as proto:
            for line in proto:
                pkg = self._PACKAGE_REGEX.match(line)
                if pkg:
                    pkg = pkg.group('package')
                    return os.path.sep.join(pkg.split('.'))
            return ''

    def _transform(self, pkg, sep, common_protos):
        """Transform to the appropriate proto package layout.

        Works with arbitrary separator (e.g., '/' for import statements,
        '.' for proto package statements, os.path.sep for filenames)
        """
        if sep != '.' and pkg.endswith('.proto'):
            dotted = pkg[:-6].replace(sep, '.')
            suffix = '.proto'
        else:
            dotted = pkg.replace(sep, '.')
            suffix = ''

        # Sanity check: Do not transform common protos.
        for common_pkg in common_protos:
            if dotted.startswith(common_pkg):
                return pkg

        # Special case: If the organization name is "google-cloud", then we
        # have to ensure that "cloud" exists in the path. The protos
        # themselves may not follow this.
        if 'cloud' not in dotted and self._organization_name == 'google-cloud':
            dotted = dotted.replace('google.', 'google.cloud.', 1)

        # Transform into the ideal proto path.
        # What essentially should happen here is that "{api}.{vN}" should
        # change to "{api}_{vN}".
        dotted = re.sub(r'\.v([\da-z_]*)([\d]+)\b', r'_v\1\2.proto', dotted)

        # Edge case: Some internal customers use "vNalpha" and "vNbeta".
        # Rather than make the regular expression more complicated, catch
        # this as a one-off.
        if re.search(r'\.v[\d]+alpha\b', dotted):
            dotted = re.sub(r'\.v([\d]+)alpha\b', r'_v\1alpha.proto', dotted)
        if re.search(r'\.v[\d]+beta\b', dotted):
            dotted = re.sub(r'\.v([\d]+)beta\b', r'_v\1beta.proto', dotted)
        if re.search(r'\.v[\d]+eap\b', dotted):
            dotted = re.sub(r'\.v([\d]+)eap\b', r'_v\1eap.proto', dotted)

        # Done; return with the appropriate separator.
        return dotted.replace('.', sep) + suffix

    def _copy_proto(self, src, dest, common_protos):
        """Copies a proto while fixing its imports"""
        with io.open(src, 'r', encoding='UTF-8') as src_lines:
            with io.open(dest, 'w+', encoding='UTF-8') as dest_file:
                for line in src_lines:
                    import_ = self._IMPORT_REGEX.match(line)
                    if import_:
                        dest_file.write('import "{}";\n'.format(
                            self._transform(
                                import_.group('package'), '/', common_protos)))
                    else:
                        dest_file.write(line)

    def _copy_and_transform_directories(
            self, src_directories, destination_directory, common_protos,
            paths=None):
        for path in src_directories:
            protos = list(protoc_utils.find_protos([path], []))
            for proto in protos:
                src_base_dirs = self._extract_base_dirs(proto)
                sub_new_src = os.path.join(
                    destination_directory,
                    self._transform(
                        src_base_dirs, os.path.sep, common_protos))
                if paths is not None:
                    paths.add(sub_new_src)

                dest = os.path.join(sub_new_src, os.path.basename(proto))
                if not os.path.exists(dest):
                    self.exec_command(['mkdir', '-p', sub_new_src])
                self._copy_proto(
                    proto, os.path.join(sub_new_src, dest), common_protos)


class PythonMoveProtosTask(task_base.TaskBase):
    default_provides = {'grpc_code_dir'}

    def execute(self, grpc_code_dir, gapic_code_dir):
        """Move the protos into the GAPIC structure.

        This copies the ``x/y/z/proto/`` directory over to be a sibling
        of ``x/y/z/gapic/`` in the GAPIC code directory. In the event of
        an inconsistency on the prefix, the GAPIC wins.

        Args:
            grpc_code_dir (str): The location where the GRPC code was
                generated.
            gapic_code_dir (str): The location where the GAPIC code was
                generated.
        """
        # Determine the appropriate source and target directory.
        # We can get this by drilling in to the GAPIC artifact until we get to
        # a "gapic" directory that is outside "docs" and "tests".
        src = self._get_proto_path(grpc_code_dir)
        target = self._get_gapic_subdir_path(gapic_code_dir)

        # Move the contents into the GAPIC directory.
        self.exec_command(['mv', src, os.path.join(target, 'proto')])

        # Create an __init__.py file in the proto directory.
        # This is necessary for Python 2.7 compatibility.
        self.exec_command([
            'touch', os.path.join(target, 'proto', '__init__.py'),
        ])

        # Remove the grpc directory.
        self.exec_command(['rm', '-rf', grpc_code_dir])

        # Clear out the grpc_code_dir, so future tasks perceive it as
        # not being a thing anymore.
        return {'grpc_code_dir': None}

    def _get_gapic_subdir_path(self, haystack):
        """Return the subpath which contains the "gapic" directory.

        Args:
            haystack (str): The top-level directory in which the subdirectory
                should appear.

        Returns:
            str: The path, relative to ``haystack``, where the subdirectory
                was found.

        Raises:
            RuntimeError: If the subdirectory is not found.
        """
        needle = 'gapic'

        for d in os.listdir(haystack):
            # Skip the 'docs' and 'tests' directories.
            # These will not contain code.
            if d == 'docs' or d == 'tests':
                continue

            for path, dirs, files in os.walk(os.path.join(haystack, d)):
                if needle in dirs:
                    return path

        raise RuntimeError('Path %s not found in %s.' % (needle, haystack))

    def _get_proto_path(self, grpc_path):
        filenames = set()
        for path, dirs, files in os.walk(grpc_path):
            filenames = filenames.union(
                {os.path.join(path, f) for f in files if f.endswith('_pb2.py')}
            )

        # The path to the protos is the common prefix of all of the proto
        # files found. This ensures that if protos reside in subpackages,
        # only the common prefix between them is used.
        if not filenames:
            raise RuntimeError(
                'Path with protobuf files not found in {0}.'.format(grpc_path),
            )
        elif len(filenames) == 1:
            # The common path for a single filename is, by definition,
            # the whole filename, not the directory name.
            return os.path.dirname(filenames.pop())
        else:
            return os.path.commonpath(list(filenames))




# Copyright 2017 Google
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, unicode_literals
import os
import subprocess
import sys

from artman.utils.logger import logger

def check_docker_requirements(docker_image):
    """Checks whether all docker-related components have been installed."""

    try:
        subprocess.check_output(['docker', '--version'])
    except OSError:
        logger.error(
            'Docker not found on path or is not installed. Refer to '
            'https://docs.docker.com/engine/installation about how to install '
            'Docker on your local machine.')
        sys.exit(128)

    try:
        output = subprocess.check_output(
            ['docker', 'images', '-q', docker_image])
        if not output:
            logger.error(
                'Cannot find artman Docker image. Run `docker pull %s` to '
                'pull the image.' % docker_image)
            sys.exit(128)
    except OSError:
        logger.error('Docker image check failed. Please file an issue at '
                    'https://github.com/googleapis/artman/issues.')
        sys.exit(128)



#!/usr/bin/env python

# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Artman smoke tests.

It generates GAPIC client libraries for Google APIs in googleapis
repo. The test will fails if any generation fails unless its artman yaml file
is on the whitelist.
"""

import argparse
import glob
import json
import logging
import os
import re
import subprocess
import sys
import yaml

from google.protobuf import json_format

from artman.config.proto import config_pb2

# Generation failure of artifact on the whitelist won't be counted as failure
# but warning. It is defined at artifact level, and one can use regex to
# whitelist the whole API or whole folder.
WHITELIST = [
    # The two configs below are failing: https://github.com/googleapis/artman/issues/483
    'python_gapic@google/devtools/containeranalysis/artman_containeranalysis_v1beta1.yaml',

    'nodejs_gapic@google/cloud/videointelligence/artman_videointelligence_v1.yaml',
    # TODO(ethanbao): Figure out the rootcause
    'java_gapic@google/cloud/dialogflow/artman_dialogflow_v2beta1_java.yaml',
    '.+@google/api/servicemanagement/artman_servicemanagement_v1.yaml',
    # TODO(pongad): talk to the API owners
    '(?:python|php)_gapic@google/api/expr/artman_cel.yaml',

    # whitelisted because of
    # java.lang.IllegalStateException: A HttpRule option must be defined.
    # (affects only PHP that generates code for JSON over HTTP)
    # https://github.com/googleapis/gapic-generator/issues/2691
    'php_gapic@google/cloud/bigquery/storage/artman_bigquerystorage_v1beta1.yaml',
    'php_gapic@google/cloud/bigquery/artman_bigquery_v2.yaml',

    # whitelisted because protos are missing Java annotations
    # necessary for client library generation. CL in flight.
    'java_gapic@google/cloud/binaryauthorization/artman_binaryauthorization_v1beta1.yaml',
    
    # TODO(hzyi): whitelisted to unblock artman release. CL coming.
    'csharp_gapic@google/firestore/artman_firestore_v1.yaml',
    'csharp_gapic@google/firestore/artman_firestore.yaml',

    # whitelisted four weird configs
    'java_grpc@google/artman_core.yaml',
    'java_grpc@google/artman_logging_external_types.yaml',
    'java_grpc@google/appengine/artman_appengine.yaml',
    'java_gapic@google/cloud/dialogflow/v2beta1/artman_dialogflow_v2beta1_java.yaml',

    # whitelisted for no HTTP annotations
    'php_gapic@google/devtools/cloudbuild/artman_cloudbuild',
    'php_gapic@google/cloud/osconfig/agentendpoint/v1beta/artman_osconfig_v1beta',

    # whitelisted because of broken GAPIC config
    # TODO(fenster): to fix that
    '.*@google/monitoring/v3/artman_monitoring.yaml',
    
    # securitycenter C# does not use gapic-generator
    'csharp_gapic@google/cloud/securitycenter/artman_securitycenter_v1p1beta1.yaml',
]

logger = logging.getLogger('smoketest')
logger.setLevel(logging.INFO)


def run_smoke_test(root_dir, user_config, log):
    log_file = _setup_logger(log)
    failure = []
    success = []
    known_fail = []
    for artman_yaml_path in glob.glob('%s/google/**/artman_*.yaml' % root_dir,
                                      recursive=True):
        artman_config = _parse(artman_yaml_path)
        # Reorder the artifacts by moving the gapic_config artifact to the end.
        # Otherwise, the following GAPIC library generation is going to use the
        # newly-generated GAPIC config instead of the one in github, leading to
        # false negative and false positive in the test result.
        artifacts = sorted(
            artman_config.artifacts,
            key=lambda a: config_pb2.Artifact.GAPIC_CONFIG == a.type)

        for artifact in artifacts:
            logger.info('Start artifact generation for %s of %s'
                        % (artifact.name, artman_yaml_path))
            if _generate_artifact(artman_yaml_path,
                                  user_config,
                                  artifact.name,
                                  root_dir,
                                  log_file):
                msg = 'Failed to generate %s of %s.' % (
                    artifact.name, artman_yaml_path)
                if _is_whitelisted(artman_yaml_path, root_dir, artifact.name):
                    known_fail.append(msg)
                else:
                    failure.append(msg)
            else:
                msg = 'Succeded to generate %s of %s.' % (
                    artifact.name, artman_yaml_path)
                success.append(msg)
            logger.info(msg)
    logger.info('================ Smoketest summary ================')
    logger.info('Success:')
    for msg in success:
        logger.info(msg)

    if known_fail:
        logger.info('Known Whitelisted Failures:')
        for msg in known_fail:
            logger.info(msg)

    if failure:
        logger.info('Failure:')
        for msg in failure:
            logger.error(msg)
        sys.exit('Smoke test failed.')


def _parse(artman_yaml_path):
    """Parse artman yaml config into corresponding protobuf."""
    with open(artman_yaml_path, 'r') as f:
        # Convert yaml into json file as protobuf python load support parsing
        # of protobuf in json or text format, not yaml.
        artman_config_json_string = json.dumps(yaml.load(f, Loader=yaml.FullLoader))
    config_pb = config_pb2.Config()
    json_format.Parse(artman_config_json_string, config_pb)

    return config_pb


def _generate_artifact(artman_config, user_config, artifact_name, root_dir, log_file):
    with open(log_file, 'a') as log:
        grpc_pipeline_args = [
            'artman',
            '--local',
            '--verbose',
            '--config', artman_config,
            '--root-dir', root_dir,
            '--user-config', user_config,
            'generate', artifact_name
        ]
        return subprocess.call(grpc_pipeline_args, stdout=log, stderr=log)


def _is_whitelisted(artman_yaml_path, root_dir, artifact_name):
    artifact = '%s@%s' % (artifact_name,
                          os.path.relpath(artman_yaml_path, root_dir))
    for whitelist in WHITELIST:
        if re.match(whitelist, artifact):
            return True
    return False


def parse_args(*args):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--root-dir',
        default='/googleapis',
        type=str,
        help='Specify where googleapis local repo lives.')
    parser.add_argument(
        '--log',
        default='/tmp/smoketest.log',
        type=str,
        help='Specify where smoketest log should be stored.')
    parser.add_argument(
        '--user-config',
        default='/artman/artman-user-config-in-docker.yaml',
        type=str,
        help='Specify where artman user config is located')
    return parser.parse_args(args=args)


def _setup_logger(log_file):
    """Setup logger with a logging FileHandler."""
    log_file_handler = logging.FileHandler(log_file, mode='a+')
    logger.addHandler(log_file_handler)
    logger.addHandler(logging.StreamHandler())
    return log_file


if __name__ == '__main__':
    flags = parse_args(*sys.argv[1:])

    run_smoke_test(os.path.abspath(flags.root_dir), os.path.abspath(flags.user_config), os.path.abspath(flags.log))



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pipelines that run GAPIC code generation."""

from __future__ import absolute_import

from artman.pipelines import code_generation as code_gen
from artman.pipelines import grpc_generation as grpc_gen
from artman.tasks import emit_success
from artman import tasks
from artman.utils import task_utils


# kwargs required by GAPIC code gen
_GAPIC_REQUIRED = ['service_yaml', 'gapic_yaml', 'language', 'aspect']

_DISCOGAPIC_REQUIRED = ['gapic_yaml', 'language']


class GapicConfigPipeline(code_gen.CodeGenerationPipelineBase):

    def __init__(self, **kwargs):
        super(GapicConfigPipeline, self).__init__(
            GapicConfigTaskFactory(), **kwargs)

class DiscoGapicConfigPipeline(code_gen.CodeGenerationPipelineBase):

    def __init__(self, **kwargs):
        super(DiscoGapicConfigPipeline, self).__init__(
            DiscoGapicConfigTaskFactory(), **kwargs)

class GapicConfigTaskFactory(code_gen.TaskFactoryBase):

    def get_tasks(self, **kwargs):
        return task_utils.instantiate_tasks([
            tasks.protoc.ProtoDescGenTask,
            tasks.gapic.GapicConfigGenTask,
            tasks.gapic.GapicConfigMoveTask
        ], kwargs)

    def get_validate_kwargs(self):
        return code_gen.COMMON_REQUIRED

    def get_invalid_kwargs(self):
        return ['language']


class DiscoGapicConfigTaskFactory(code_gen.TaskFactoryBase):

    def get_tasks(self, **kwargs):
        return task_utils.instantiate_tasks([
            tasks.gapic.DiscoGapicConfigGenTask,
            tasks.gapic.GapicConfigMoveTask
        ], kwargs)

    def get_validate_kwargs(self):
        return code_gen.COMMON_DISCO_REQUIRED

    def get_invalid_kwargs(self):
        return ['language']


class GapicOnlyClientPipeline(code_gen.CodeGenerationPipelineBase):
    """The pipeline for generating a GAPIC_ONLY layer, other layers
    (like PROTOBUF or GRPC) will not be generated.
    """
    def __init__(self, language, **kwargs):
        super(GapicOnlyClientPipeline, self).__init__(
            GapicOnlyTaskFactory(),
            language=language,
            **kwargs
        )


class GapicClientPipeline(code_gen.CodeGenerationPipelineBase):
    """The pipeline for generating a complete GAPIC.

    This is intended to be the only command that needs to run to generate
    a complete GAPIC.
    """
    def __init__(self, language, **kwargs):
        super(GapicClientPipeline, self).__init__(
            GapicTaskFactory(),
            language=language,
            **kwargs
        )


class DiscoGapicClientPipeline(code_gen.CodeGenerationPipelineBase):
    """The pipeline for generating a complete GAPIC from a Discovery document.

    This is intended to be the only command that needs to run to generate
    a complete GAPIC.
    """
    def __init__(self, language, **kwargs):
        super(DiscoGapicClientPipeline, self).__init__(
            DiscoGapicTaskFactory(),
            language=language,
            **kwargs
        )

class CSharpPackagingTaskFactory(code_gen.TaskFactoryBase):

    def get_tasks(self, **kwargs):
        return [
            tasks.gapic.CSharpGapicPackagingTask
        ]

    def get_validate_kwargs(self):
        return ['gapic_code_dir', 'grpc_code_dir', 'proto_code_dir', 'gapic_yaml']

    def get_invalid_kwargs(self):
        return []


PACKAGING_TASK_FACTORY_DICT = {
    'csharp': CSharpPackagingTaskFactory
}

class GapicTaskFactory(code_gen.TaskFactoryBase):
    """A task factory describing GAPIC generation tasks.

    Language specific tasks may be defined in language-specific methods
    on this class.
    """
    def get_tasks(self, **kwargs):
        """Return the full list of instantiated tasks to generate a GAPIC.

        Args:
            kwargs (dict): Keyword arguments, which are passed through.
        """
        # Note: For languages where generating a GAPIC implies also generating
        # the GRPC or Proto library, we still generate the GAPIC _first_, even
        # though the GRPC is the dependency. This is because in some languages,
        # the GRPC is "tucked into" the GAPIC, and that process is much easier
        # if the code is generated in this order.
        answer = []

        if 'gapic_code_dir' in kwargs:
            answer = self._get_gapic_codegen_tasks(**kwargs)

        for grpc_task in self._get_grpc_codegen_tasks(**kwargs):
            if grpc_task not in answer:
                answer.append(grpc_task)

        for packaging_task in self._get_packaging_tasks(**kwargs):
            if packaging_task not in answer:
                answer.append(packaging_task)

        answer += emit_success.TASKS
        return task_utils.instantiate_tasks(answer, kwargs)

    def _get_gapic_codegen_tasks(self, language, **kwargs):
        """Return the code generation tasks necessary for creating a GAPIC.

        Args:
            language (str): The language.

        Returns:
            list: A list of Task subclasses.
        """
        return [
            tasks.protoc.ProtoDescGenTask,
            tasks.descriptor.get_descriptor_set_task(language),
            tasks.package_metadata.PackageMetadataConfigGenTask,
            tasks.gapic.GapicCodeGenTask,
            tasks.format.get_format_task(language),
        ]

    def _get_grpc_codegen_tasks(self, language, **kw):
        """Return the code generation tasks for making a GRPC package.

        Args:
            language (str): The language code is being generated in.
            kw (dict): Additional keyword arguments passed through to the
                grpc codegen task factory.

        Returns:
            list: A list of Task subclasses defined by the GRPC task factory.
        """

        grpc_factory = grpc_gen.ProtoGenTaskFactory(gen_grpc=True,
                                                    language=language, **kw)
        return grpc_factory.get_grpc_codegen_tasks(**kw)

    def _get_packaging_tasks(self, language, **kw):
        """Return the code generation tasks for packaging

        Args:
            language (str): The language code is being generated in.
            kw (dict): Additional keyword arguments passed through to the
                proto codegen task factory.

        Returns:
            list: A list of Task subclasses defined by the packaging task factory.
        """
        if language in PACKAGING_TASK_FACTORY_DICT:
            packaging_factory = PACKAGING_TASK_FACTORY_DICT[language]()
            return packaging_factory.get_tasks()
        return []

    def get_validate_kwargs(self):
        return _GAPIC_REQUIRED + code_gen.COMMON_REQUIRED

    def get_invalid_kwargs(self):
        return []


class DiscoGapicTaskFactory(code_gen.TaskFactoryBase):
    """A task factory describing GAPIC generation tasks on Discovery docs.

    Language specific tasks may be defined in language-specific methods
    on this class.
    """
    def get_tasks(self, **kwargs):
        """Return the full list of instantiated tasks to generate a GAPIC.

        Args:
            kwargs (dict): Keyword arguments, which are passed through.
        """
        answer = []

        if 'gapic_code_dir' in kwargs:
            answer = self._get_gapic_codegen_tasks(**kwargs)

        for packaging_task in self._get_packaging_tasks(**kwargs):
            if packaging_task not in answer:
                answer.append(packaging_task)

        answer += emit_success.TASKS
        return task_utils.instantiate_tasks(answer, kwargs)

    def _get_gapic_codegen_tasks(self, language, **kwargs):
        """Return the code generation tasks necessary for creating a GAPIC
        from a Discovery document.

        Args:
            language (str): The language.

        Returns:
            list: A list of Task subclasses.
        """
        return [
            tasks.io.PrepareOutputDirectoryTask,
            tasks.package_metadata.PackageMetadataConfigGenTask,
            tasks.gapic.DiscoGapicCodeGenTask,
            tasks.format.get_format_task(language),
        ]

    def _get_packaging_tasks(self, language, **kw):
        """Return the code generation tasks for packaging

        Args:
            language (str): The language code is being generated in.
            kw (dict): Additional keyword arguments passed through to the
                proto codegen task factory.

        Returns:
            list: A list of Task subclasses defined by the packaging task factory.
        """
        if language in PACKAGING_TASK_FACTORY_DICT:
            packaging_factory = PACKAGING_TASK_FACTORY_DICT[language]()
            return packaging_factory.get_tasks()
        return []

    def get_validate_kwargs(self):
        return _DISCOGAPIC_REQUIRED + code_gen.COMMON_DISCO_REQUIRED

    def get_invalid_kwargs(self):
        return []


class GapicOnlyTaskFactory(GapicTaskFactory):
    """A task factory describing GAPIC_ONLY generation tasks."""
    def _get_grpc_codegen_tasks(self, **kw):
        return[]

    def _get_packaging_tasks(self, **kw):
        return []



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions related to tasks"""

import os.path
import re
import subprocess

import six


def get_java_tool_path(toolkit_path, tool_name):
    path = os.path.join(toolkit_path, 'build', 'toolpaths', tool_name)
    if not os.path.exists(path):
        run_gradle_task(toolkit_path, 'createToolPaths')
    return path


def gapic_gen_task(toolkit_path, task_args):
    toolkit_path = os.path.realpath(os.path.expanduser(toolkit_path))
    gapic_jar = os.path.join(toolkit_path, 'build/libs/gapic-generator-latest-fatjar.jar')
    if not os.path.exists(gapic_jar):
        run_gradle_task(toolkit_path, 'fatJar')
    return ['java', '-cp', gapic_jar, 'com.google.api.codegen.GeneratorMain'] + task_args


def run_gradle_task(toolkit_path, task_name, task_args=()):
    """Generates a command for a gradle task."""
    toolkit_path = os.path.realpath(os.path.expanduser(toolkit_path))
    subprocess.check_output([os.path.join(toolkit_path, 'gradlew'), '-p', toolkit_path,
            task_name, '-Pclargs=' + ','.join(task_args)])


def api_full_name(api_name, api_version, organization_name):
    """Canonical full name for an API; used to generate output directories and
    package name"""
    if api_version:
        return '-'.join([organization_name, api_name, api_version])
    else:
        return '-'.join([organization_name, api_name])


def is_output_gcloud(language, gapic_code_dir):
    """Check if the gapic_code_dir is a part of gcloud project.
    """
    gapic_code_dir = os.path.abspath(gapic_code_dir)
    if language == 'nodejs':
        cloud_dirname = 'google-cloud-node'
    else:
        cloud_dirname = 'google-cloud-' + language
    if gapic_code_dir.find(os.path.sep + cloud_dirname + os.path.sep) >= 0:
        return True
    return re.search(os.path.sep + 'gcloud-[a-z]+' + os.path.sep,
                     gapic_code_dir)


def instantiate_tasks(task_class_list, inject):
    """Instantiates a list of Tasks. Generates a name for each task based on
    the class name, and if available the language and api_name settings.
    """
    tasks = []
    for task_class in task_class_list:
        name = task_class.__name__
        if inject.get('language'):
            name += '-' + inject['language']
        if inject.get('api_name'):
            name += '-' + inject['api_name']
        if inject.get('api_version'):
            name += '-' + inject['api_version']
        tasks.append(task_class(name, inject=inject))
    return tasks



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The new artman CLI with the following syntax.

    artman [Options] generate <artifact_name>

.. note::
    Only local execution is supported at this moment. The CLI syntax is
    beta, and might have changes in the future.
"""

from __future__ import absolute_import
from logging import INFO
import argparse
from distutils.dir_util import copy_tree
import io
import os
import pprint
import subprocess
import sys
import tempfile
import traceback

import pkg_resources
from ruamel import yaml
from taskflow import engines

from artman.config import converter, loader
from artman.config.proto.config_pb2 import Artifact, Config
from artman.config.proto.user_config_pb2 import UserConfig
from artman.cli import support
from artman.pipelines import pipeline_factory
from artman.utils import config_util
from artman.utils.logger import logger, setup_logging

VERSION = pkg_resources.get_distribution('googleapis-artman').version
ARTMAN_DOCKER_IMAGE = 'googleapis/artman:%s' % VERSION
RUNNING_IN_ARTMAN_DOCKER_TOKEN = 'RUNNING_IN_ARTMAN_DOCKER'
DEFAULT_OUTPUT_DIR = './artman-genfiles'

def main(*args):
    """Main method of artman."""
    # If no arguments are sent, we are using the entry point; derive
    # them from sys.argv.
    if not args:
        args = sys.argv[1:]

    # Get to a normalized set of arguments.
    flags = parse_args(*args)
    user_config = loader.read_user_config(flags.user_config)
    _adjust_root_dir(flags.root_dir)
    pipeline_name, pipeline_kwargs = normalize_flags(flags, user_config)

    if flags.local:
        try:
            pipeline = pipeline_factory.make_pipeline(pipeline_name,
                                                      **pipeline_kwargs)
            # Hardcoded to run pipeline in serial engine, though not necessarily.
            engine = engines.load(
                pipeline.flow, engine='serial', store=pipeline.kwargs)
            engine.run()
        except:
            logger.error(traceback.format_exc())
            sys.exit(32)
        finally:
            _change_owner(flags, pipeline_name, pipeline_kwargs)
    else:
        support.check_docker_requirements(flags.image)
        # Note: artman currently won't work if input directory doesn't contain
        # common-protos.
        logger.info('Running artman command in a Docker instance.')
        _run_artman_in_docker(flags)


def _adjust_root_dir(root_dir):
    """Adjust input directory to use versioned common config and/or protos.

    Some common protos will be needed during protoc compilation, but are not
    provided by users in some cases. When such shared proto directories are
    not provided, we copy and use the versioned ones.
    """
    if os.getenv(RUNNING_IN_ARTMAN_DOCKER_TOKEN):
        # Only doing this when running inside Docker container
        common_proto_dirs = [
            'google/api',
            'google/iam/v1',
            'google/longrunning',
            'google/rpc',
            'google/type',
        ]
        # /googleapis is the root of the versioned googleapis repo
        # inside Artman Docker image.
        for src_dir in common_proto_dirs:
            if not os.path.exists(os.path.join(root_dir, src_dir)):
                copy_tree(os.path.join('/googleapis', src_dir),
                          os.path.join(root_dir, src_dir))


def parse_args(*args):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {0}'.format(VERSION),
    )
    parser.add_argument(
        '--config',
        type=str,
        default='artman.yaml',
        help='[Optional] Specify path to artman config yaml, which can be '
        'either an absolute path, or a path relative to the input '
        'directory (specified by `--root-dir` flag). Default to '
        '`artman.yaml`', )
    parser.add_argument(
        '--output-dir',
        type=str,
        default=DEFAULT_OUTPUT_DIR,
        help='[Optional] Directory to store output generated by artman. '
        'Default to `' + DEFAULT_OUTPUT_DIR + '`', )
    parser.add_argument(
        '--root-dir',
        type=str,
        default='',
        help='[Optional] Directory with all input that is needed by artman, '
        'which includes but is not limited to API protos, service config yaml '
        'and GAPIC config yaml. It will be passed to protobuf compiler via '
        '`-I` flag in order to generate descriptor. If not specified, it '
        'will use the current working directory.',
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='store_const',
        const=10,
        default=None,
        dest='verbosity',
        help='Show verbose / debug output.', )
    parser.add_argument(
        '--user-config',
        default='~/.artman/config.yaml',
        help='[Optional] User configuration file to stores credentials like '
        'GitHub credentials. Default to `~/.artman/config.yaml`', )
    parser.add_argument(
        '--local',
        dest='local',
        action='store_true',
        help='[Optional] If specified, running the artman on the local host '
        'machine instead of artman docker instance that have all binaries '
        'installed. Note: one will have to make sure all binaries get '
        'installed on the local machine with this flag, a full list can '
        'be found at '
        'https://github.com/googleapis/artman/blob/master/Dockerfile', )
    parser.set_defaults(local=False)
    parser.add_argument(
        '--image',
        default=ARTMAN_DOCKER_IMAGE,
        help=('[Optional] Specify docker image used by artman when running in '
              'a Docker instance. Default to `%s`' % ARTMAN_DOCKER_IMAGE))
    parser.add_argument(
        '--generator-args',
        type=str,
        default=None,
        help='Additional arguments to pass to gapic-generator')


    # Add sub-commands.
    subparsers = parser.add_subparsers(
        dest='subcommand', help='Support [generate] sub-commands')

    # `generate` sub-command.
    parser_generate = subparsers.add_parser(
        'generate', help='Generate artifact')
    parser_generate.add_argument(
        'artifact_name',
        type=str,
        help='[Required] Name of the artifact for artman to generate. Must '
        'match an artifact in the artman config yaml.')
    parser_generate.add_argument(
        '--aspect',
        type=str,
        default=None,
        help='[Optional] Aspect of output to generate: ALL, CODE, or PACKAGE')

    return parser.parse_args(args=args)


def normalize_flags(flags, user_config):
    """Combine the argparse flags and user configuration together.

    Args:
        flags (argparse.Namespace): The flags parsed from sys.argv
        user_config (dict): The user configuration taken from
                            ~/.artman/config.yaml.

    Returns:
        tuple (str, dict): 2-tuple containing:
            - pipeline name
            - pipeline arguments
    """
    if flags.root_dir:
        flags.root_dir = os.path.abspath(flags.root_dir)
        flags.config = os.path.join(flags.root_dir, flags.config)
    else:
        flags.root_dir = os.getcwd()
        flags.config = os.path.abspath(flags.config)
    root_dir = flags.root_dir
    flags.output_dir = os.path.abspath(flags.output_dir)
    pipeline_args = {}

    # Determine logging verbosity and then set up logging.
    verbosity = INFO
    if getattr(flags, 'verbosity', None):
        verbosity = getattr(flags, 'verbosity')
    setup_logging(verbosity)

    # Save local paths, if applicable.
    # This allows the user to override the path to api-client-staging or
    # toolkit on his or her machine.
    pipeline_args['root_dir'] = root_dir
    pipeline_args['toolkit_path'] = user_config.local.toolkit
    pipeline_args['generator_args'] = flags.generator_args

    artman_config_path = flags.config
    if not os.path.isfile(artman_config_path):
        logger.error(
            'Artman config file `%s` doesn\'t exist.' % artman_config_path)
        sys.exit(96)

    try:
        artifact_config = loader.load_artifact_config(
            artman_config_path, flags.artifact_name, flags.aspect)
    except ValueError as ve:
        logger.error('Artifact config loading failed with `%s`' % ve)
        sys.exit(96)

    legacy_config_dict = converter.convert_to_legacy_config_dict(
        artifact_config, root_dir, flags.output_dir)
    logger.debug('Below is the legacy config after conversion:\n%s' %
                 pprint.pformat(legacy_config_dict))

    language = Artifact.Language.Name(
        artifact_config.language).lower()

    # Set the pipeline
    artifact_type = artifact_config.type
    pipeline_args['artifact_type'] = Artifact.Type.Name(artifact_type)
    pipeline_args['aspect'] = Artifact.Aspect.Name(artifact_config.aspect)
    if artifact_type == Artifact.GAPIC_ONLY:
        pipeline_name = 'GapicOnlyClientPipeline'
        pipeline_args['language'] = language
    elif artifact_type == Artifact.GAPIC:
        pipeline_name = 'GapicClientPipeline'
        pipeline_args['language'] = language
    elif artifact_type == Artifact.DISCOGAPIC:
        pipeline_name = 'DiscoGapicClientPipeline'
        pipeline_args['language'] = language
        pipeline_args['discovery_doc'] = artifact_config.discovery_doc
    elif artifact_type == Artifact.GRPC:
        pipeline_name = 'GrpcClientPipeline'
        pipeline_args['language'] = language
    elif artifact_type == Artifact.GAPIC_CONFIG:
        pipeline_name = 'GapicConfigPipeline'
    elif artifact_type == Artifact.DISCOGAPIC_CONFIG:
        pipeline_name = 'DiscoGapicConfigPipeline'
        pipeline_args['discovery_doc'] = artifact_config.discovery_doc
        if os.path.abspath(flags.output_dir) != os.path.abspath(DEFAULT_OUTPUT_DIR):
            logger.warning("`output_dir` is ignored in DiscoGapicConfigGen. "
             + "Yamls are saved at the path specified by `gapic_yaml`.")
        pipeline_args['output_dir'] = tempfile.mkdtemp()
    elif artifact_type == Artifact.PROTOBUF:
        pipeline_name = 'ProtoClientPipeline'
        pipeline_args['language'] = language
    else:
        raise ValueError('Unrecognized artifact.')

    # Parse out the full configuration.
    config_args = config_util.load_config_spec(legacy_config_dict, language)
    config_args.update(pipeline_args)
    pipeline_args = config_args
    # Print out the final arguments to stdout, to help the user with
    # possible debugging.
    pipeline_args_repr = yaml.dump(
        pipeline_args,
        block_seq_indent=2,
        default_flow_style=False,
        indent=2, )
    logger.info('Final args:')
    for line in pipeline_args_repr.split('\n'):
        if 'token' in line:
            index = line.index(':')
            line = line[:index + 2] + '<< REDACTED >>'
        logger.info('  {0}'.format(line))

    # Return the final arguments.
    return pipeline_name, pipeline_args

def _run_artman_in_docker(flags):
    """Executes artman command.

    Args:
        root_dir: The input directory that will be mounted to artman docker
            container as local googleapis directory.
    Returns:
        The output directory with artman-generated files.
    """
    ARTMAN_CONTAINER_NAME = 'artman-docker'
    root_dir = flags.root_dir
    output_dir = flags.output_dir
    artman_config_dirname = os.path.dirname(flags.config)
    docker_image = flags.image

    inner_artman_cmd_str = ' '.join(["'" + arg + "'" for arg in sys.argv[1:]])
    # Because artman now supports setting root dir in either command line or
    # user config, make sure `--root-dir` flag gets explicitly passed to the
    # artman command running inside Artman Docker container.
    if '--root-dir' not in inner_artman_cmd_str:
      inner_artman_cmd_str = '--root-dir %s %s' % (
          root_dir, inner_artman_cmd_str)

    # TODO(ethanbao): Such folder to folder mounting won't work on windows.
    base_cmd = [
        'docker', 'run', '--name', ARTMAN_CONTAINER_NAME, '--rm', '-i', '-t',
        '-e', 'HOST_USER_ID=%s' % os.getuid(),
        '-e', 'HOST_GROUP_ID=%s' % os.getgid(),
        '-e', '%s=True' % RUNNING_IN_ARTMAN_DOCKER_TOKEN,
        '-v', '%s:%s' % (root_dir, root_dir),
        '-v', '%s:%s' % (output_dir, output_dir),
        '-v', '%s:%s' % (artman_config_dirname, artman_config_dirname),
        '-w', root_dir
    ]
    base_cmd.extend([docker_image, '/bin/bash', '-c'])

    inner_artman_debug_cmd_str = inner_artman_cmd_str
    # Because debug_cmd is run inside the Docker image, we want to
    # make sure --local is set
    if '--local' not in inner_artman_debug_cmd_str:
        inner_artman_debug_cmd_str = '--local %s' % inner_artman_debug_cmd_str
    debug_cmd = list(base_cmd)
    debug_cmd.append('"artman %s; bash"' % inner_artman_debug_cmd_str)

    cmd = base_cmd
    cmd.append('artman --local %s' % (inner_artman_cmd_str))
    try:
        output = subprocess.check_output(cmd)
        logger.info(output.decode('utf8'))
        return output_dir
    except subprocess.CalledProcessError as e:
        logger.error(e.output.decode('utf8'))
        logger.error(
            'Artman execution failed. For additional logging, re-run the '
            'command with the "--verbose" flag')
        sys.exit(32)
    finally:
        logger.debug('For further inspection inside docker container, run `%s`'
                     % ' '.join(debug_cmd))


def _change_owner(flags, pipeline_name, pipeline_kwargs):
    """Change file/directory ownership if necessary."""
    user_host_id = int(os.getenv('HOST_USER_ID', 0))
    group_host_id = int(os.getenv('HOST_GROUP_ID', 0))
    # When artman runs in Docker instance, all output files are by default
    # owned by `root`, making it non-editable by Docker host user. When host
    # user id and group id get passed through environment variables via
    # Docker `-e` flag, artman will change the owner based on the specified
    # user id and group id.
    if not user_host_id or not group_host_id:
        return
    # Change ownership of output directory.
    if os.path.exists(flags.output_dir):
        _change_directory_owner(flags.output_dir, user_host_id, group_host_id)

    # Change the local repo directory if specified.
    if 'local_repo_dir' in pipeline_kwargs:
        local_repo_dir = pipeline_kwargs['local_repo_dir']
        if (os.path.exists(local_repo_dir)):
            _change_directory_owner(local_repo_dir, user_host_id, group_host_id)

    if pipeline_kwargs['gapic_yaml']:
        gapic_config_path = pipeline_kwargs['gapic_yaml']
        if (os.path.exists(gapic_config_path) and
            ('GapicConfigPipeline' == pipeline_name
                or 'DiscoGapicConfigPipeline' == pipeline_name)):
            # There is a trick that the gapic config output is generated to
            # input directory, where it is supposed to be in order to be
            # used as an input for other artifact generation. With that
            # the gapic config output is not located in the output directory,
            # but the input directory. Make the explicit chown in this case.
            os.chown(gapic_config_path, user_host_id, group_host_id)


def _change_directory_owner(directory, user_host_id, group_host_id):
    """Change ownership recursively for everything under the given directory."""
    for root, dirs, files in os.walk(directory):
        os.chown(root, user_host_id, group_host_id)
        for d in dirs:
            os.chown(
                os.path.join(root, d), user_host_id, group_host_id)
        for f in files:
            os.chown(
                os.path.join(root, f), user_host_id, group_host_id)

if __name__ == "__main__":
    main()



# Copyright 2017 Google
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from collections import OrderedDict
import getpass
import importlib
import io
import json
import logging
import os

import six

from ruamel import yaml
import stringcase
import yaml

from artman.config.proto.user_config_pb2 import UserConfig, LocalConfig, GitHubConfig
from artman.utils.logger import logger
from artman.utils.logger import setup_logging
from google.protobuf.json_format import MessageToJson

__all__ = ('configure',)


def configure(log_level=logging.INFO):
    """Allow the user to write a new configuration file.

    Returns:
        int: An exit status.
    """
    user_config = UserConfig()

    # Walk the user through basic configuration.
    setup_logging(log_level)
    logger.info('Welcome to artman. We will get you configured.')
    logger.info('When this is done, config will be stored in ~/.artman/config.yaml.')
    logger.info('')

    # Go through each step.
    # These are split out to make testing them easier.
    user_config.local.CopyFrom(_configure_local_config())

    try:
        config_dir = os.path.expanduser('~/.artman/')
        os.makedirs(config_dir)
    except OSError:
        pass
    _write_pb_to_yaml(user_config, os.path.join(config_dir, 'config.yaml'))
    logger.success('Configuration written successfully to '
                   '~/.artman/config.yaml.')


def _configure_local_config():
    """Determine and return artman user local config.

    Returns:
        LocalConfig: The new artman local config settings.
    """
    answer = LocalConfig()

    # Ask the user for a local toolkit location.
    while not answer.toolkit:
        location = six.moves.input(
            'Where is your local toolkit repository? (If you do not have a '
            'toolkit repository, clone https://github.com/googleapis/toolkit/) '
            'Please provide an absolute path: ')
        if location:
            answer.toolkit = location

    # Done; return the answer.
    return answer


def _write_pb_to_yaml(pb, output):
    # Add yaml representer so that yaml dump can dump OrderedDict. The code
    # is coming from https://stackoverflow.com/questions/16782112.
    yaml.add_representer(OrderedDict, _represent_ordereddict)

    json_obj = _order_dict(json.loads(MessageToJson(pb)))
    with open(output, 'w') as outfile:
        yaml.dump(json_obj, outfile, default_flow_style=False)


def _represent_ordereddict(dumper, data):
    value = []
    for item_key, item_value in data.items():
        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)

        value.append((node_key, node_value))

    return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)


def _order_dict(od):
    # The whole key order is based on the simple name of the config field
    # instead of its full name (`username` vs `github.username`).
    keyorder = [
        'local', 'toolkit', 'github', 'username', 'token'
    ]
    res = OrderedDict()
    for k, v in sorted(od.items(), key=lambda i: keyorder.index(i[0])):
        if isinstance(v, dict):
            res[k] = _order_dict(v)
        elif isinstance(v, list):
            if isinstance(v[0], dict):
                result = []
                for d2 in v:
                    result.append(_order_dict(d2))
                res[k] = result
            else:
                res[k] = v
        else:
            res[k] = v
    return res



# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, unicode_literals
import os

from artman.tasks import task_base
from artman.utils.logger import logger


class EmitSuccess(task_base.TaskBase):
    """Emit a success message with the gapic_code_dir location."""

    def execute(self, gapic_code_dir, grpc_code_dir=None):
        """Emit a success message. No publishing is being done.

        Args:
            gapic_code_dir (str): The current GAPIC code location.
            grpc_code_dir (str): The current GRPC code location, if any.
        """
        userhome = os.path.expanduser('~')
        gapic_loc = os.path.realpath(gapic_code_dir).replace(userhome, '~')
        logger.success('Code generated: {0}'.format(gapic_loc))
        if grpc_code_dir:
            grpc_loc = os.path.realpath(grpc_code_dir).replace(userhome, '~')
            logger.success('GRPC code generated: {0}'.format(grpc_loc))


TASKS = (EmitSuccess,)
